"""Dataloaders for tabular data"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/10_dataloaders/12_tabular_dataloaders.ipynb.

# %% auto 0
__all__ = ['XYDataLoader', 'MultiShapeLoader']

# %% ../../nbs/10_dataloaders/12_tabular_dataloaders.ipynb 3
import logging
logging.basicConfig(level=logging.INFO)

import numpy as np
from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Literal
import pandas as pd
import math

from .base import BaseDataLoader

from sklearn.preprocessing import StandardScaler, MinMaxScaler

# %% ../../nbs/10_dataloaders/12_tabular_dataloaders.ipynb 4
class XYDataLoader(BaseDataLoader):

    """

    A class for datasets with the typicall X, Y structure. Both X
    and Y are numpy arrays. X may be of shape (datapoints, features) or (datapoints, sequence_length, features) 
    if lag features are used. The prep_lag_features can be used to create those lag features. Y is of shape
    (datapoints, units).

    """
    
    def __init__(self,
        X: np.ndarray,
        Y: np.ndarray,
        val_index_start: Union[int, None] = None, 
        test_index_start: Union[int, None] = None, 
        lag_window_params: Union[dict] = None, # default: {'lag_window': 0, 'include_y': False, 'pre_calc': False}
        normalize_features: Union[dict] = None, # default: {'normalize': True, 'ignore_one_hot': True}
    ):

        self.X = X
        self.Y = Y

        self.val_index_start = val_index_start
        self.test_index_start = test_index_start

        # train index ends either at the start of the validation set, the start of the test set or at the end of the dataset
        if self.val_index_start is not None:
            self.train_index_end = self.val_index_start-1
        elif self.test_index_start is not None:
            self.train_index_end = self.test_index_start-1
        else:
            self.train_index_end = len(Y)-1

        self.dataset_type = "train"

        normalize_features = normalize_features or {'normalize': True, 'ignore_one_hot': True}
        lag_window_params = lag_window_params or {'lag_window': 0, 'include_y': False, 'pre_calc': False}

        self.normalize_features(**normalize_features, initial_normalization=True)
        self.prep_lag_features(**lag_window_params)

        # X must at least have datapoint and feature dimension
        if len(X.shape) == 1:
            self.X = X.reshape(-1, 1)
        
        # Y must have at least datapoint and unit dimension (even if only one unit is present)
        if len(Y.shape) == 1:
            self.Y = Y.reshape(-1, 1)

        assert len(X) == len(Y), 'X and Y must have the same length'

        self.num_units = Y.shape[1] # shape 0 is alsways time, shape 1 is the number of units (e.g., SKUs)

        super().__init__()

    def normalize_features(self,
        normalize: bool = True,
        ignore_one_hot: bool = True,
        initial_normalization=False # Flag if it is set before having added lag features
        ):

        """
        Normalize features using a standard scaler. If ignore_one_hot is true, one-hot encoded features are not normalized.

        """

        if normalize:

            scaler = StandardScaler()

            if initial_normalization:

                if len(self.X.shape) == 3:
                    raise ValueError('Normalization not possible with lag features. Please set initial_normalization=False')
            
                scaler.fit(self.X[:self.train_index_end+1]) # +1 to include the last training point
                scaler.transform(self.X)

                if initial_normalization:
                    return
                else:
                    raise NotImplementedError('Normalization after lag features have been set not implemented yet')

                    # Idea:
                        # remove time dimension
                        # normalize features
                        # add time_dimension back
                    # Problem:
                        # usage of prep_lag_features needs to ensure y is not added a second time

    def prep_lag_features(self,
        lag_window: int = 0, # length of the lage window
        include_y: bool = False, # if lag demand shall be included as feature
        pre_calc: bool = False # if all lags are pre-calculated for the entire dataset
        ):

        """
        Create lag feature for the dataset. If "inlcude_y" is true, then a lag-1 of of the target variable is added as a feature.
        If lag-window is > 0, the lag features are added as middle dimension to X. Note that this, e.g., means that with a lag
        window of 1, the data will include 2 time steps, the current features including lag-1 demand and the lag-1 features
        including lag-2 demand. If pre-calc is true, all these calculations are performed on the entire dataset reduce
        computation time later on at the expense of increases memory usage. 

        """
        # to be discussed: Do we need option to only provide lag demand wihtout lag features?
        self.lag_window = lag_window
        self.pre_calc = pre_calc
        self.include_y = include_y
        
        if self.pre_calc:
            if self.include_y:
                # add additional column to X with demand shifted by 1
                self.X = np.concatenate((self.X, np.roll(self.Y, 1, axis=0)), axis=1)
                self.X = self.X[1:] # remove first row
                self.Y = self.Y[1:] # remove first row
                
                self.val_index_start = self.val_index_start-1
                self.test_index_start = self.test_index_start-1
                self.train_index_end  = self.train_index_end-1
        
            if self.lag_window is not None and self.lag_window > 0:

                # add lag features as dimention 2 to X (making it dimension (datapoints, sequence_length, features))
                X_lag = np.zeros((self.X.shape[0], self.lag_window+1, self.X.shape[1]))
                for i in range(self.lag_window+1):
                    if i == 0:
                        features = self.X
                    else:    
                        features = self.X[:-i, :]
                    X_lag[i:, self.lag_window-i, :] = features
                self.X = X_lag[self.lag_window:]
                self.Y = self.Y[self.lag_window:]

                self.val_index_start = self.val_index_start-self.lag_window
                self.test_index_start = self.test_index_start-self.lag_window
                self.train_index_end  = self.train_index_end-self.lag_window

        else:
            self.lag_window = None
            self.include_y = False
            # add time dimension to X

    def update_lag_features(self,
        lag_window: int,
        ):

        """ Update lag window parameters for dataloader object that is already initialized """

        raise NotImplementedError('Not implemented yet')

        # Problem: updating lag_features naively would shorten the dataset each time it is called

    def __getitem__(self, idx): 

        """ get item by index, depending on the dataset type (train, val, test)"""

        if self.dataset_type == "train":
            if idx > self.train_index_end:
                raise IndexError(f'index {idx} out of range{self.train_index_end}')
            idx = idx

        elif self.dataset_type == "val":
            idx = idx + self.val_index_start
            
            if idx >= self.test_index_start:
                raise IndexError(f'index{idx} out of range{self.test_index_start}')
            
        elif self.dataset_type == "test":
            idx = idx + self.test_index_start
            
            if idx >= len(self.X):
                raise IndexError(f'index{idx} out of range{len(self.X)}')
        
        else:
            raise ValueError('dataset_type not set')

        return self.X[idx], self.Y[idx]

    def __len__(self):
        return len(self.X)
    
    @property
    def X_shape(self):
        return self.X.shape
    
    @property
    def Y_shape(self):
        return self.Y.shape

    @property
    def len_train(self):
        return self.train_index_end+1

    @property
    def len_val(self):
        if self.val_index_start is None:
            raise ValueError('no validation set defined')
        return self.test_index_start-self.val_index_start

    @property
    def len_test(self):
        if self.test_index_start is None:
            raise ValueError('no test set defined')
        return len(self.Y)-self.test_index_start

    def get_all_X(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        """
        Returns the entire features dataset.
        Return either the train, val, test, or all data.
        """

        if dataset_type == 'train':
            return self.X[:self.val_index_start].copy() if self.X is not None else None
        elif dataset_type == 'val':
            return self.X[self.val_index_start:self.test_index_start].copy() if self.X is not None else None
        elif dataset_type == 'test':
            return self.X[self.test_index_start:].copy() if self.X is not None else None
        elif dataset_type == 'all':
            return self.X.copy() if self.X is not None else None
        else:
            raise ValueError('dataset_type not recognized')

    def get_all_Y(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        """
        Returns the entire target dataset.
        Return either the train, val, test, or all data.
        """

        if dataset_type == 'train':
            return self.Y[:self.val_index_start].copy() if self.Y is not None else None
        elif dataset_type == 'val':
            return self.Y[self.val_index_start:self.test_index_start].copy() if self.Y is not None else None
        elif dataset_type == 'test':
            return self.Y[self.test_index_start:].copy() if self.Y is not None else None
        elif dataset_type == 'all':
            return self.Y.copy() if self.Y is not None else None
        else:
            raise ValueError('dataset_type not recognized')
        

# %% ../../nbs/10_dataloaders/12_tabular_dataloaders.ipynb 18
class MultiShapeLoader(BaseDataLoader):

    """
    A class designed for comlex datasets with mutlipe feature types. The class is more
    memory-efficient than the XYDataLoader, as it separate the storeage of SKU-specific
    feature, time-specific features, and time-SKU-specific features. The class works generically
    as long as those feature classes are provided during pre-processing. The class is designed 
    to handle classic learning, but able to work in a meta-learning pipeline where no SKU-dimension
    is present and the model needs to make prediction on SKU-time level without knowhing the
    specific SKU.
    """
    
    def __init__(self,
        # mandatory data
        demand: pd.DataFrame, # Demand data of shape time x SKU
        time_features: pd.DataFrame, # Features constant over SKU of shape time x time_features
        time_SKU_features: pd.DataFrame, # Features varying over time and SKU of shape time x (time_SKU_features*SKU) with double index
        
        # optional data
        mask: pd.DataFrame = None, # Mask of shape time x SKU telling which SKUs are available at which time (can be used as mask during trainig or added to features)
        SKU_features: pd.DataFrame = None, # Features constant over time of shape SKU x SKU_features - only for algorithms learning across SKUs
        
        val_index_start: Union[int, None] = None, # Validation index start on the time dimension
        test_index_start: Union[int, None] = None, # Test index start on the time dimension
        in_sample_val_test_SKUs: List = None, # SKUs in the training set to be used for validation and testing, out-of-sample w.r.t. time dimension
        out_of_sample_val_SKUs: List = None, # SKUs to be hold-out for validation (can be same as test if no validation on out-of-sample SKUs required)
        out_of_sample_test_SKUs: List = None, # SKUs to be hold-out for testing
        lag_window_params: dict | None = None, # default: {'lag_window': 0, 'include_y': False, 'pre_calc': True}
        normalize_features: dict | None = None, # default: {'normalize': True, 'ignore_one_hot': True}
        engineered_SKU_features: Union[dict] = None, # default: ["mean_demand", "std_demand", "kurtosis_demand", "skewness_demand", "percentile_10_demand", "percentile_30_demand", "median_demand", "percentile_70_demand", "percentile_90_demand", "inter_quartile_range"]
        use_engineered_SKU_features: bool = False, # if engineered features shall be used
        include_non_available: bool = False, # if timestep/SKU combination where the SKU was not available for sale shall be included. If included, it will be used as feature, otherwise as mask.
        train_subset: int = None ,# if only a subset of SKUs is used for training. Will always contain in_sample_val_test_SKUs and then fills the rest with random SKUs
        train_subset_SKUs: List = None, # if train_subset is set, specific SKUs can be provided
        meta_learn_units: bool = False, # if units (SKUs) are trained in the batch dimension to meta-learn across SKUs
        lag_demand_normalization: Literal['minmax', 'standard', 'no_normalization'] | None = "standard", # minmax, standard, no_normalization or None. If None, same demand_normalization
        demand_normalization: Literal['minmax', 'standard', 'no_normalization'] = 'no_normalization', # 'standard' or 'minmax'
        demand_unit_size: float | None = None, # use same convention as for other dataloaders and enviornments, but here only full decimal values are allowed
        provide_additional_target: bool = False, # follows ICL convention by providing actual demand to token, with the last token receiving 0
        permutate_inputs: bool = False, # if the inputs shall be permutated during training for meta-learning
    ):
     
        logging.info("Setting main env attributes")
        # Set data
        self.demand = demand
        self.SKU_features = SKU_features
        self.SKU_features_out_of_sample_val, self.SKU_features_out_of_sample_test = None, None # will be overwritten if relevant
        self.time_features = time_features
        self.time_SKU_features = time_SKU_features
        self.mask = mask
        self.permutate_inputs = permutate_inputs

        # convert dtypes to float
        self.demand = self.demand.astype(float)
        self.time_features = self.time_features.astype(float)
        self.time_SKU_features = self.time_SKU_features.astype(float)
        if self.SKU_features is not None:
            self.SKU_features = self.SKU_features.astype(float)
        if self.mask is not None:
            self.mask = self.mask.astype(float)

        # Set default values for dict inputs:
        normalize_features = normalize_features or {'normalize': True, 'ignore_one_hot': True}
        lag_window_params = lag_window_params or {'lag_window': 0, 'include_y': False, 'pre_calc': False}
        self.lag_window_params = lag_window_params # lag window parameters saved as attribute
        engineered_SKU_features = engineered_SKU_features or ["mean_demand", "std_demand", "kurtosis_demand", "skewness_demand", "percentile_10_demand", "percentile_30_demand", "median_demand", "percentile_70_demand", "percentile_90_demand", "inter_quartile_range"]
        if not use_engineered_SKU_features:
            engineered_SKU_features = None

        # Set further training parameters
        self.include_non_available = include_non_available
        self.meta_learn_units = meta_learn_units
        train_subset, train_subset_SKUs = self.set_train_subset(train_subset, train_subset_SKUs) # set the attributes train_subset and train_subset_SKUs
        self.lag_demand_normalization = lag_demand_normalization if lag_demand_normalization is not None else demand_normalization
        self.demand_normalization = demand_normalization
        if demand_unit_size is not None:
            self.demand_unit_size = int(-math.log10(demand_unit_size))
            if self.demand_unit_size % 1 != 0:
                raise ValueError('demand_unit_size must be a full decimal value')
            else:
                self.demand_unit_size = int(self.demand_unit_size)
        else:
            self.demand_unit_size = None
        self.provide_additional_target = provide_additional_target
        
        # Some necessary flags
        ## Whether the features are already normalized
        self.normalized_in_sample_SKUs = False
        self.normalized_out_of_sample_SKUs = False
        self.added_engineereed_features_to_in_sample_SKUs = False
        self.added_engineereed_features_to_out_of_sample_val_SKUs = False
        self.added_engineereed_features_to_out_of_sample_test_SKUs = False
        ## What data to return in the __getitem__ method
        self.SKU_type = "in_sample" # or "out_of_sample_val" or "out_of_sample_test" # What kind of SKU retruning in __getitem__
        self.dataset_type = "train" # or "val" or "test", affecting the time-dimension

        logging.info("Setting indices for validation and test set")
        # Determine train, val, test indices (time dimension only!)
        self.val_index_start = val_index_start
        self.test_index_start = test_index_start
        self.train_index_start = self.lag_window_params["lag_window"] # start index for training data later to allow for lag features
        self.train_index_start += self.lag_window_params["include_y"] # if lag demand is included as feature need one more timestep

        # train index ends either at the start of the validation set, the start of the test set or at the end of the dataset
        if self.val_index_start is not None:
            self.train_index_end = self.val_index_start-1
        elif self.test_index_start is not None:
            self.train_index_end = self.test_index_start-1
        else:
            self.train_index_end = len(self.demand)-1

        logging.info("Setting out-of-sample SKUs")
        # set out-of-sample SKUs - note that this is not done by indecing, but the SKUs data will be
        # removed from the in-sample data and stored in separate attributes. Therefore, there are no
        # index attributes for out-of-sample SKUs. The feature attributes are stored by the function below.
        self.out_of_sample_val_SKUs, self.out_of_sample_test_SKUs = self.set_out_of_sample_SKUs(out_of_sample_val_SKUs, out_of_sample_test_SKUs)

        if self.out_of_sample_val_SKUs is not None and self.out_of_sample_test_SKUs is not None:
            if len(self.out_of_sample_val_SKUs) != 0 or len(self.out_of_sample_test_SKUs) != 0:
                self.out_of_sample = True
            else:
                self.out_of_sample = False
        else:
            self.out_of_sample = False

        logging.info("Identifying training and in-sample validation and test SKUs")
        self.in_sample_val_test_SKUs = self.set_in_sample_val_test_SKUs(in_sample_val_test_SKUs) # need to determine index later after out-of-sample units are removed
        self.train_SKUs = self.identify_train_SKUs(train_subset, train_subset_SKUs) # need to determine index later after out-of-sample units are removed 
        if self.in_sample_val_test_SKUs is None:
            self.in_sample_val_test_SKUs = self.train_SKUs # validation and training SKUs are teh same

        logging.info("Determine number of units and features")
        # Num units is relevant for the output dimension when validating and testing. If the model is not trained as a 
        # meta learner it is identicall in traing, and validation/testing. If it is trained as a meta learner, the output
        # dimension is the number of in_sample_val_test_SKUs, iregardless of the number of SKUs in the training set.
        self.num_units = len(self.in_sample_val_test_SKUs) if self.in_sample_val_test_SKUs is not None else len(self.demand.columns)
        
        # Determine number of features
        self.num_time_SKU_features = len(self.time_SKU_features.columns.get_level_values(0).unique())
        if lag_window_params["include_y"]:
            self.num_time_SKU_features += 1 # lag demand is time and SKU specific
        if include_non_available:
            self.num_time_SKU_features += 1 # non-availability is time and SKU specific
        if self.provide_additional_target:
            self.num_time_SKU_features += 1
        self.num_time_features = len(self.time_features.columns)
        self.num_SKU_features = len(self.SKU_features.columns) if self.SKU_features is not None else 0
        if engineered_SKU_features is not None:
            self.num_SKU_features += len(engineered_SKU_features) # engineered features are SKU specific

        # print("xxxxxxx Number features per type:")
        # print("num_time_SKU_features:", self.num_time_SKU_features)
        # print("num_time_features:", self.num_time_features)
        # print("num_SKU_features:", self.num_SKU_features)

        # Determine total number of features per datapoint
        self.num_features = self.num_SKU_features + self.num_time_features + self.num_time_SKU_features

        if engineered_SKU_features is not None:
            logging.info("Creating engineered SKU features for training data")
            engineered_SKU_features_data = self.build_engineered_SKU_features(engineered_SKU_features, self.demand.iloc[:self.train_index_end+1]) # only for training data initially
            self.SKU_features = pd.concat([self.SKU_features, engineered_SKU_features_data.transpose()], axis=1)
            self.added_engineereed_features_to_in_sample_SKUs = True

            logging.info("Creating engineered SKU features for out-of-sample validation data")
            engineered_SKU_features_data = self.build_engineered_SKU_features(engineered_SKU_features, self.demand_out_of_sample_val) # only for out-of-sample validation data
            self.SKU_features_out_of_sample_val = pd.concat([self.SKU_features_out_of_sample_val, engineered_SKU_features_data.transpose()], axis=1)
            self.added_engineereed_features_to_out_of_sample_val_SKUs = True

            logging.info("Creating engineered SKU features for out-of-sample test data")
            engineered_SKU_features_data = self.build_engineered_SKU_features(engineered_SKU_features, self.demand_out_of_sample_test) # only for out-of-sample test data
            self.SKU_features_out_of_sample_test = pd.concat([self.SKU_features_out_of_sample_test, engineered_SKU_features_data.transpose()], axis=1)
            self.added_engineereed_features_to_out_of_sample_test_SKUs = True

        logging.info("Normalizing in-sample SKU features (based on training timesteps)")
        self.normalize_demand_and_features_in_sample(normalize=normalize_features["normalize"], ignore_one_hot=normalize_features["ignore_one_hot"],initial_normalization=True)

        if self.out_of_sample:
            logging.info("Normalizing out-of-sample SKU features")
            self.normalize_demand_and_features_out_of_sample(normalize=normalize_features["normalize"], ignore_one_hot=normalize_features["ignore_one_hot"],initial_normalization=True)

        logging.info("Saving data as numpy and saving indices")
        # store row and column indices of demand, SKU_features time_features mask and then convert to numpy array
        logging.info("--Saving indices")

        ############ in sample data ############
        self.in_sample_val_test_SKUs_indices = self.demand.columns.get_indexer(self.in_sample_val_test_SKUs) if self.in_sample_val_test_SKUs is not None else None

        self.train_SKUs_indices = self.demand.columns.get_indexer(self.train_SKUs)

        self.demand_indices = self.save_indices(self.demand)
        self.SKU_features_indices = self.save_indices(self.SKU_features) if self.SKU_features is not None else None
        self.time_features_indices = self.save_indices(self.time_features)
        self.time_SKU_features_indices = self.save_indices(self.time_SKU_features)
        self.mask_indices = self.save_indices(self.mask)

        logging.info("--Converting to numpy - in sample")
        self.demand = self.demand.to_numpy()
        self.demand_lag = self.demand_lag.to_numpy()
        self.SKU_features = self.SKU_features.to_numpy() if self.SKU_features is not None else None
        self.time_features = self.time_features.to_numpy()
        self.time_SKU_features = self.time_SKU_features.to_numpy()
        self.mask = self.mask.to_numpy()

        ############ out of sample data ############

        if self.out_of_sample:
            self.out_of_sample_val_SKUs_indices = self.demand_out_of_sample_val.columns.get_indexer(self.out_of_sample_val_SKUs)
            self.out_of_sample_test_SKUs_indices = self.demand_out_of_sample_test.columns.get_indexer(self.out_of_sample_test_SKUs)

            self.demand_out_of_sample_val_indices = self.save_indices(self.demand_out_of_sample_val)
            self.SKU_features_out_of_sample_val_indices = self.save_indices(self.SKU_features_out_of_sample_val) if self.SKU_features_out_of_sample_val is not None else None
            self.time_SKU_features_out_of_sample_val_indices = self.save_indices(self.time_SKU_features_out_of_sample_val)
            self.mask_out_of_sample_val_indices = self.save_indices(self.mask_out_of_sample_val)
        
            logging.info("--Converting to numpy - out of sample val")
            self.demand_out_of_sample_val = self.demand_out_of_sample_val.to_numpy()
            self.demand_lag_out_of_sample_val = self.demand_lag_out_of_sample_val.to_numpy()
            self.SKU_features_out_of_sample_val = self.SKU_features_out_of_sample_val.to_numpy() if self.SKU_features_out_of_sample_val is not None else None
            self.time_SKU_features_out_of_sample_val = self.time_SKU_features_out_of_sample_val.to_numpy()
            self.mask_out_of_sample_val = self.mask_out_of_sample_val.to_numpy()

            self.demand_out_of_sample_test_indices = self.save_indices(self.demand_out_of_sample_test)
            self.SKU_features_out_of_sample_test_indices = self.save_indices(self.SKU_features_out_of_sample_test) if self.SKU_features_out_of_sample_test is not None else None
            self.time_SKU_features_out_of_sample_test_indices = self.save_indices(self.time_SKU_features_out_of_sample_test)
            self.mask_out_of_sample_test_indices = self.save_indices(self.mask_out_of_sample_test)

            logging.info("--Converting to numpy - out of sample test")
            self.demand_out_of_sample_test = self.demand_out_of_sample_test.to_numpy()
            self.demand_lag_out_of_sample_test = self.demand_lag_out_of_sample_test.to_numpy()
            self.SKU_features_out_of_sample_test = self.SKU_features_out_of_sample_test.to_numpy() if self.SKU_features_out_of_sample_test is not None else None
            self.time_SKU_features_out_of_sample_test = self.time_SKU_features_out_of_sample_test.to_numpy()
            self.mask_out_of_sample_test = self.mask_out_of_sample_test.to_numpy()

        

        ############ final params ############
        self.len_train_time = self.train_index_end-self.train_index_start+1

        print("len_train_time:", self.len_train_time)
        print("num_units:", self.num_units)
        print("len train_SKU_indices:", len(self.train_SKUs_indices))

        if self.meta_learn_units:
            logging.info("--Creating time-SKU index for training data")
            self.sku_time_index = [(i, j) for i in self.train_SKUs_indices for j in range(self.len_train_time)]

        self.set_return_sku("in_sample")

        super().__init__()

    def set_train_subset(self, train_subset, train_subset_SKUs):
        """ Prepare setting the attributes train_subset and train_subset_SKUs """

        if train_subset is not None and train_subset < 1:
            raise ValueError('train_subset must be an integer larger than 0')

        if train_subset_SKUs is not None and not isinstance(train_subset_SKUs, list):
            raise ValueError(f'train_subset_SKUs must be a list of SKUs, got {type(train_subset_SKUs)}')

        if train_subset_SKUs is not None:
            if train_subset is None:
                logging.info("--Infering train_subset from train_subset_SKUs")
                train_subset = len(train_subset_SKUs)
            else:
                if train_subset != len(train_subset_SKUs):
                    raise ValueError('train_subset_SKUs must have the same length as train_subset')
        
        return train_subset, train_subset_SKUs

    def set_in_sample_val_test_SKUs(self, in_sample_val_test_SKUs):

        if in_sample_val_test_SKUs is None:
            in_sample_val_test_SKUs_indices = None
        else:
            if not self.meta_learn_units:
                raise ValueError('in_sample_val_test_SKUs can only be set if meta_learn_units is True/n \
                otherwise the output dimension needs to be the same for training and validation/testing')
            if not set(in_sample_val_test_SKUs).issubset(self.demand.columns):

                missing_SKUs = 0
                names = []
                for i in in_sample_val_test_SKUs:
                    if i not in self.demand.columns:
                        missing_SKUs += 1
                        names.append(i)
                
                print(names)
            

                raise ValueError('in_sample_val_test_SKUs must be a subset of all SKUs')

        return in_sample_val_test_SKUs
            
    def identify_train_SKUs(self, train_subset, train_subset_SKUs):
        """ determine which SKUs are used for training, validation and testing """
        
        if train_subset is not None:

            if train_subset_SKUs is not None:
                # check that all train_SKUs are in demand.collumns
                if not set(train_subset_SKUs).issubset(self.demand.columns):
                    raise ValueError('train_subset_SKUs must be a subset of all training SKUs')
                if self.in_sample_val_test_SKUs is not None:
                    if not set(self.in_sample_val_test_SKUs).issubset(train_subset_SKUs):
                        raise ValueError('train_subset_SKUs must contain in_sample_val_test_SKUs')
            else:
                if self.in_sample_val_test_SKUs is not None and train_subset < len(self.in_sample_val_test_SKUs):
                    raise ValueError(f'train_subset ({train_subset}) must be equal or larger than the number of in_sample_val_test_SKUs ({len(self.in_sample_val_test_SKUs)})')
                train_subset_SKUs = self.in_sample_val_test_SKUs if self.in_sample_val_test_SKUs is not None else []
                remaining_SKUs = self.demand.columns.difference(train_subset_SKUs)
                additional_SKUs = np.random.choice(remaining_SKUs, train_subset-len(train_subset_SKUs), replace=False)
                train_SKUs = np.concatenate((train_subset_SKUs, additional_SKUs))
                # order train_SKUs as in demand.columns
                train_subset_SKUs = [sku for sku in self.demand.columns if sku in train_SKUs]
        else:
            train_subset_SKUs = self.demand.columns # val and test SKUs have been removed before, only training SKUs remain 

        return train_subset_SKUs
    
    def set_out_of_sample_SKUs(self,
            out_of_sample_val_SKUs: List | np.ndarray | None, # SKUs for validation
            out_of_sample_test_SKUs: List | np.ndarray | None, # SKUs for testing. If same dataset shall be used, it can be identical
            ):

        """
        Sets out-of-sample SKUs for validation and test datasets, and removes them from the in-sample data.
        It handles the cases that no out-of-sample SKUs are set (both inputs are None), one of them is set,
        both of them are set to different values, and both of them are identical. In-sample data remain,
        while out-of-sample data are moved to seperate attributes.

        """

        # Note: one of the two might be None, both might be None, or both might be identical

        self.test_out_of_sample_SKUs(out_of_sample_val_SKUs, out_of_sample_test_SKUs)

        if out_of_sample_val_SKUs is None:
            out_of_sample_val_SKUs = []
        if out_of_sample_test_SKUs is None:
            out_of_sample_test_SKUs = []

        for sku, attr_suffix in [(out_of_sample_val_SKUs, 'val'), (out_of_sample_test_SKUs, 'test')]:
            if sku is not None:
                logging.info(f"--Setting out-of-sample {attr_suffix} SKUs")
                # Set attributes for out-of-sample SKUs
                setattr(self, f'demand_out_of_sample_{attr_suffix}', self.demand.loc[:, sku])
                if self.SKU_features is not None:
                    setattr(self, f'SKU_features_out_of_sample_{attr_suffix}', self.SKU_features.loc[sku])
                setattr(self, f'time_SKU_features_out_of_sample_{attr_suffix}', # here SKU are in columns on index level 2
                        self.time_SKU_features.loc[:, pd.IndexSlice[:, sku]]) 
                setattr(self, f'mask_out_of_sample_{attr_suffix}', self.mask.loc[:, sku])
                # time_features are independent of SKU, so no need to set them

        # unique values of uniion of both lists
        skus_to_remove = list(set(out_of_sample_val_SKUs).union(out_of_sample_test_SKUs))
        if skus_to_remove:
            logging.info(f"--Removing {len(skus_to_remove)} out-of-sample SKUs from in-sample data")
            # Remove out-of-sample SKUs from in-sample data
            self.demand.drop(columns=skus_to_remove, inplace=True)
            if self.SKU_features is not None:
                self.SKU_features.drop(index=skus_to_remove, inplace=True)
            for single_sku in skus_to_remove:
                columns_to_drop = self.time_SKU_features.columns.get_loc_level(single_sku, level=1)
                self.time_SKU_features.drop(columns=self.time_SKU_features.columns[columns_to_drop[0]], inplace=True)
            self.mask.drop(columns=skus_to_remove, inplace=True)

        return out_of_sample_val_SKUs, out_of_sample_test_SKUs

    def test_out_of_sample_SKUs(self, out_of_sample_val_SKUs, out_of_sample_test_SKUs):
        
        """
        Validates the out-of-sample SKU lists for compatibility. Ensures that the inputs are either None, lists, or numpy arrays,
        and that they meet specific conditions regarding uniqueness and overlap.
        """

        # Check if inputs are either None, list, or numpy array
        if not (isinstance(out_of_sample_val_SKUs, (list, np.ndarray)) or out_of_sample_val_SKUs is None):
            raise ValueError("out_of_sample_val_SKUs must be a list, numpy array, or None.")
        if not (isinstance(out_of_sample_test_SKUs, (list, np.ndarray)) or out_of_sample_test_SKUs is None):
            raise ValueError("out_of_sample_test_SKUs must be a list, numpy array, or None.")

        # Convert None to empty list for easier processing
        if out_of_sample_val_SKUs is None:
            out_of_sample_val_SKUs = []
        if out_of_sample_test_SKUs is None:
            out_of_sample_test_SKUs = []

        # Ensure all entries are unique within each list
        if len(out_of_sample_val_SKUs) != len(set(out_of_sample_val_SKUs)):
            raise ValueError("out_of_sample_val_SKUs contains duplicate entries.")
        if len(out_of_sample_test_SKUs) != len(set(out_of_sample_test_SKUs)):
            raise ValueError("out_of_sample_test_SKUs contains duplicate entries.")

        # Ensure there are no overlapping SKUs between validation and test sets (unless they are meant to be identical)

        if out_of_sample_val_SKUs is not None and out_of_sample_test_SKUs is not None:
            overlap = set(out_of_sample_val_SKUs).intersection(set(out_of_sample_test_SKUs))

            if len(overlap) > 0:
                if len(overlap) == len(out_of_sample_val_SKUs):
                    logging.warning("--Out of sample validation and test SKUs are identical - ensure this is intended.")
                else:
                    raise ValueError(f"Validation and test SKUs must not overlap. Overlapping SKUs: {overlap}")

        logging.debug("--Out-of-sample SKUs have passed validation checks.")
        
    @staticmethod
    def build_engineered_SKU_features(engineered_SKU_features: List, demand: pd.DataFrame):

        """
        Create engineered features for each SKU
        """

        feature_names = []
        feature_values = []

        for feature in engineered_SKU_features:

            if feature == "mean_demand":
                mean_demand = demand.mean(axis=0)
            elif feature == "std_demand":
                std_demand = demand.std(axis=0)
            elif feature == "kurtosis_demand":
                kurtosis_demand = demand.kurtosis(axis=0)
            elif feature == "skewness_demand":
                skewness_demand = demand.skew(axis=0)
            elif feature == "percentile_10_demand":
                percentile_10_demand = demand.quantile(0.1, axis=0)
            elif feature == "percentile_30_demand":
                percentile_30_demand = demand.quantile(0.3, axis=0)
            elif feature == "median_demand":
                median_demand = demand.median(axis=0)
            elif feature == "percentile_70_demand":
                percentile_70_demand = demand.quantile(0.7, axis=0)
            elif feature == "percentile_90_demand":
                percentile_90_demand = demand.quantile(0.9, axis=0)
            elif feature == "inter_quartile_range":
                inter_quartile_range = demand.quantile(0.75, axis=0) - demand.quantile(0.25, axis=0)
            else:  
                raise ValueError(f'Feature {feature} not recognized')
            
            feature_names.append(feature)
            feature_values.append(locals()[feature])

        return pd.DataFrame(feature_values, columns=demand.columns, index=feature_names)
    
    def normalize_demand_and_features_in_sample(self,
        normalize: bool = True,
        ignore_one_hot: bool = True,
        initial_normalization = False # Flag if it is set before having added lag features
        ):

        """
        Normalize features using a standard scaler. If ignore_one_hot is true, one-hot encoded features are not normalized.
        """

        if normalize:

            if self.normalized_in_sample_SKUs:
                raise ValueError('Features already normalized')

            if self.demand_normalization == 'minmax':
                self.scaler_demand = MinMaxScaler()
                self.scaler_out_of_sample_val_demand = MinMaxScaler()
                self.scaler_out_of_sample_test_demand = MinMaxScaler()
            elif self.demand_normalization == 'standard':
                self.scaler_demand = StandardScaler()
                self.scaler_out_of_sample_val_demand = StandardScaler()
                self.scaler_out_of_sample_test_demand = StandardScaler()
            elif self.demand_normalization == 'no_normalization':
                self.scaler_demand = None
                self.scaler_out_of_sample_val_demand = None
                self.scaler_out_of_sample_test_demand = None
            else:
                raise ValueError('demand_normalization must be either "minmax", "standard", or "no_normalization"')
            
            # If demand data used for lag feautures is normalized differently
            if self.lag_demand_normalization != self.demand_normalization:
                if self.lag_demand_normalization == 'minmax':
                    self.scaler_demand_lag = MinMaxScaler()
                    self.scaler_out_of_sample_val_demand_lag = MinMaxScaler()
                    self.scaler_out_of_sample_test_demand_lag = MinMaxScaler()
                elif self.lag_demand_normalization == 'standard':
                    self.scaler_demand_lag = StandardScaler()
                    self.scaler_out_of_sample_val_demand_lag = StandardScaler()
                    self.scaler_out_of_sample_test_demand_lag = StandardScaler()
                elif self.lag_demand_normalization == 'no_normalization':
                    self.scaler_demand_lag = None
                    self.scaler_out_of_sample_val_demand_lag = None
                    self.scaler_out_of_sample_test_demand_lag = None
                else:
                    raise ValueError('lag_demand_normalization must be either "minmax", "standard", or "no_normalization"')
            
            self.scaler_SKU_features = StandardScaler() if self.SKU_features is not None else None # only one since out of sample uses the same fit on known  skus
            self.scaler_time_features = StandardScaler() # only one since time-features are shared between in-sample and out-of-sample SKUs
            self.scaler_time_SKU_features= [StandardScaler() for _ in range(self.num_time_SKU_features)]
            self.scaler_out_of_sample_val_SKU_features = [StandardScaler() for _ in range(self.num_time_SKU_features)]
            self.scaler_out_of_sample_test_SKU_features = [StandardScaler() for _ in range(self.num_time_SKU_features)]

            if initial_normalization:

                logging.info("--Normalizing demand")

                # Prepare data for lag_demand
                self.demand_lag = self.demand.copy() # store original demand values for lag demand

                # Normalize demand targets
                if self.demand_normalization != 'no_normalization':
                    # Normalizing per SKU on time dimension
                    self.scaler_demand.fit(self.demand[:self.train_index_end+1])
                    transformed_demand = self.scaler_demand.transform(self.demand)
                    self.demand.iloc[:,:] = transformed_demand

                # Set unit size for demand targets
                if self.demand_unit_size != None:
                    self.demand = np.round(self.demand, self.demand_unit_size)

                # If separate normalization for lag demand, normalize it
                if self.lag_demand_normalization != self.demand_normalization:
                    if self.lag_demand_normalization != 'no_normalization':
                        self.demand_lag = self.demand.copy() # if lag demand shall be normalized, build on the normalized demand (to account for slight variations due to rounding)
                        self.scaler_demand_lag.fit(self.demand_lag[:self.train_index_end+1])
                        transformed_demand_lag = self.scaler_demand_lag.transform(self.demand_lag)
                        self.demand_lag.iloc[:,:] = transformed_demand_lag
                
                # If lag demand shall be normalized the same way, then copy the normalized demand
                else:
                    self.demand_lag = self.demand.copy()

                if self.SKU_features is not None:
                    logging.info("--Normalizing SKU features")
                    # Normalizing across SKUs, no time dimension present
                    continuous_features = [col for col in self.SKU_features.columns if not self.is_one_hot(self.SKU_features[col])] if ignore_one_hot else self.SKU_features.columns
                    self.SKU_features_to_fit = continuous_features
                    if len(continuous_features) > 0:
                        self.scaler_SKU_features.fit(self.SKU_features[continuous_features]) # SKU features are already calculated based on training index
                        transformed_SKU_features = self.scaler_SKU_features.transform(self.SKU_features[continuous_features])
                        self.SKU_features[continuous_features] = transformed_SKU_features

                logging.info("--Normalizing time features")
                # Normalizting time features (no SKU dimension)
                continuous_features = [col for col in self.time_features.columns if not self.is_one_hot(self.time_features[col])] if ignore_one_hot else self.time_features.columns
                self.time_features_to_fit = continuous_features
                if len(continuous_features) > 0:
                    self.scaler_time_features.fit(self.time_features.loc[:, continuous_features].iloc[:self.train_index_end+1, :])
                    transformed_time_features = self.scaler_time_features.transform(self.time_features.loc[:,continuous_features])
                    self.time_features.loc[:,continuous_features] = transformed_time_features

                logging.info("--Normalizing time-SKU features")
                # Normalize time-SKU features (double-indexed)
                self.time_SKU_features_to_fit = dict()
                for i, feature in enumerate(self.time_SKU_features.columns.get_level_values(0).unique()):
                    # Select all columns corresponding to the current feature in level 0
                    feature_df = self.time_SKU_features.xs(key=feature, axis=1, level=0)

                    should_scale = not ignore_one_hot or (not self.is_one_hot_across_skus(feature_df) if ignore_one_hot else False)
                    self.time_SKU_features_to_fit.update({feature: should_scale})
                    if should_scale:
                        self.scaler_time_SKU_features[i].fit(feature_df[:self.train_index_end+1])
                        transformed_feature_df = self.scaler_time_SKU_features[i].transform(feature_df)
                        self.time_SKU_features.loc[:, (feature, slice(None))] = transformed_feature_df

            
                self.normalized_in_sample_SKUs = True

            else:
                raise NotImplementedError('Training data can only normalized during initialization - later normlization not implemented yet')

    def normalize_demand_and_features_out_of_sample(self,
        normalize: bool = True,
        ignore_one_hot: bool = True,
        initial_normalization = False # Flag if it is set before having added lag features
        ):

        """
        Normalize features using a standard scaler. If ignore_one_hot is true, one-hot encoded features are not normalized.
        """

        if self.out_of_sample_val_SKUs is None and self.out_of_sample_test_SKUs is None:
            return

        if normalize:

            if self.normalized_out_of_sample_SKUs:
                raise ValueError('Features already normalized')

            if initial_normalization:

                logging.info("--Normalizing demand")

                # Prepare data for lag_demand

                self.demand_lag_out_of_sample_test = self.demand_out_of_sample_test.copy()
                self.demand_lag_out_of_sample_val = self.demand_out_of_sample_val.copy()

                # Normalize demand targets
                if self.demand_normalization != 'no_normalization':
                    # Normalizing per SKU on time dimension

                    self.scaler_out_of_sample_test_demand.fit(self.demand_out_of_sample_test[:self.train_index_end+1])
                    transformed_demand = self.scaler_out_of_sample_test_demand.transform(self.demand_out_of_sample_test)
                    self.demand_out_of_sample_test.iloc[:,:] = transformed_demand

                    self.scaler_out_of_sample_val_demand.fit(self.demand_out_of_sample_val[:self.train_index_end+1])
                    transformed_demand = self.scaler_out_of_sample_val_demand.transform(self.demand_out_of_sample_val)
                    self.demand_out_of_sample_val.iloc[:,:] = transformed_demand
                
                # Set unit size for demand targets
                if self.demand_unit_size != None:
                    self.demand_out_of_sample_test = np.round(self.demand_out_of_sample_test, self.demand_unit_size)
                    self.demand_out_of_sample_val = np.round(self.demand_out_of_sample_val, self.demand_unit_size)

                # If separate normalization for lag demand, normalize it
                if self.lag_demand_normalization != self.demand_normalization:
                    if self.lag_demand_normalization != 'no_normalization':
                        
                        self.demand_lag_out_of_sample_test = self.demand_out_of_sample_test.copy()
                        
                        self.demand_lag_out_of_sample_test = self.demand_out_of_sample_test.copy()
                        self.scaler_out_of_sample_test_demand_lag.fit(self.demand_lag_out_of_sample_test[:self.train_index_end+1])
                        transformed_demand_lag = self.scaler_out_of_sample_test_demand_lag.transform(self.demand_lag_out_of_sample_test)
                        self.demand_lag_out_of_sample_test.iloc[:,:] = transformed_demand_lag
                        
                        self.demand_lag_out_of_sample_val = self.demand_out_of_sample_val.copy()
                        self.demand_lag_out_of_sample_val = self.demand_out_of_sample_val.copy()
                        self.scaler_out_of_sample_val_demand_lag.fit(self.demand_lag_out_of_sample_val[:self.train_index_end+1])
                        transformed_demand_lag = self.scaler_out_of_sample_val_demand_lag.transform(self.demand_lag_out_of_sample_val)
                        self.demand_lag_out_of_sample_val.iloc[:,:] = transformed_demand_lag
                    
                # If lag demand shall be normalized the same way, then copy the normalized demand
                else:
                    self.demand_lag_out_of_sample_test = self.demand_out_of_sample_test.copy()
                    self.demand_lag_out_of_sample_val = self.demand_out_of_sample_val.copy()
                    
                if self.SKU_features is not None:
                    logging.info("--Normalizing SKU features")
                    # Normalizing across SKUs, no time dimension present
                    continuous_features = self.SKU_features_to_fit
                    if len(continuous_features) > 0:
                        transformed_SKU_features = self.scaler_SKU_features.transform(self.SKU_features_out_of_sample_test[continuous_features])
                        self.SKU_features_out_of_sample_test[continuous_features] = transformed_SKU_features
                    
                    if len(continuous_features) > 0:
                        transformed_SKU_features = self.scaler_SKU_features.transform(self.SKU_features_out_of_sample_val[continuous_features])
                        self.SKU_features_out_of_sample_val[continuous_features] = transformed_SKU_features

                logging.info("--Normalizing time-SKU features")
                
                for i, feature in enumerate(self.time_SKU_features_out_of_sample_test.columns.get_level_values(0).unique()):
                    # Select all columns corresponding to the current feature in level 0
                    feature_df = self.time_SKU_features_out_of_sample_test.xs(key=feature, axis=1, level=0)

                    should_scale = self.time_SKU_features_to_fit[feature]
                    if should_scale:
                        self.scaler_out_of_sample_test_SKU_features[i].fit(feature_df[:self.train_index_end+1])
                        transformed_feature_df = self.scaler_out_of_sample_test_SKU_features[i].transform(feature_df)
                        self.time_SKU_features_out_of_sample_test.loc[:, (feature, slice(None))] = transformed_feature_df
                
                for i, feature in enumerate(self.time_SKU_features_out_of_sample_val.columns.get_level_values(0).unique()):
                    # Select all columns corresponding to the current feature in level 0
                    feature_df = self.time_SKU_features_out_of_sample_val.xs(key=feature, axis=1, level=0)

                    should_scale = self.time_SKU_features_to_fit[feature]
                    if should_scale:
                        self.scaler_out_of_sample_val_SKU_features[i].fit(feature_df[:self.train_index_end+1])
                        transformed_feature_df = self.scaler_out_of_sample_val_SKU_features[i].transform(feature_df)
                        self.time_SKU_features_out_of_sample_val.loc[:, (feature, slice(None))] = transformed_feature_df
            
                self.normalized_out_of_sample_SKUs = True

            else:
                raise NotImplementedError('Training data can only normalized during initialization - later normlization not implemented yet')

    def update_lag_features(self,
        lag_window: int,
        ):

        """ Update lag window parameters for dataloader object that is already initialized """

        raise NotImplementedError('Not implemented yet')

        # Problem: updating lag_features naively would shorten the dataset each time it is called

    def get_time_SKU_idx(self, idx):

        """ get time and SKU index by index, depending on the dataset type (train, val, test) """

        # print("mode of dataloader when getting item:", self.dataset_type)

        if self.dataset_type == "train":

            if self.meta_learn_units:

                if idx >= len(self.sku_time_index):
                    raise IndexError(f'index {idx} out of range{len(self.sku_time_index)}')
                idx_sku, idx_time, = self.sku_time_index[idx]
                idx_skus = [idx_sku]

            else:
                if idx+self.train_index_start > self.train_index_end:
                    raise IndexError(f'index {idx} out of range{self.train_index_end-self.train_index_start}')
                idx_skus = self.train_SKUs_indices
                idx_time = idx
            idx_time += self.train_index_start

        elif self.dataset_type == "val":
            idx_time = idx + self.val_index_start
            if self.return_SKU_type == "in_sample":
                if self.in_sample_val_test_SKUs is not None:
                    idx_skus = self.in_sample_val_test_SKUs_indices
                else:
                    idx_skus = self.train_SKUs_indices
            elif self.return_SKU_type == "out_of_sample_val":
                idx_skus = self.out_of_sample_val_SKUs_indices
            elif self.return_SKU_type == "out_of_sample_test":
                idx_skus = self.out_of_sample_test_SKUs_indices
            else:
                raise ValueError('return_SKU_type not set')
            
            if idx >= self.test_index_start:
                raise IndexError(f'index{idx} out of range{self.test_index_start}')
        
        elif self.dataset_type == "test":
            idx_time = idx + self.test_index_start
            if self.return_SKU_type == "in_sample":
                if self.in_sample_val_test_SKUs is not None:
                    idx_skus = self.in_sample_val_test_SKUs_indices
                else:
                    idx_skus = self.train_SKUs_indices
            elif self.return_SKU_type == "out_of_sample_val":
                idx_skus = self.out_of_sample_val_SKUs_indices
            elif self.return_SKU_type == "out_of_sample_test":
                idx_skus = self.out_of_sample_test_SKUs_indices
            else:
                raise ValueError('return_SKU_type not set')
            
            if idx >= len(self.demand):
                raise IndexError(f'index{idx} out of range{len(self.demand)}')
        else:
            raise ValueError('dataset_type not set')

        return idx_time, idx_skus

    def __getitem__(self, idx: int):

        """ get item by index, depending on the dataset type (train, val, test)"""

        if self.dataset_type != "train":

            time_features = self.time_features # time features independent of SKU

            if self.return_SKU_type == "in_sample":
                demand = self.demand
                demand_lag = self.demand_lag
                SKU_features = self.SKU_features
                time_SKU_features = self.time_SKU_features
                mask = self.mask
                len_SKUs = len(self.train_SKUs)
            elif self.return_SKU_type == "out_of_sample_val":
                demand = self.demand_out_of_sample_val
                demand_lag = self.demand_lag_out_of_sample_val
                SKU_features = self.SKU_features_out_of_sample_val
                time_SKU_features = self.time_SKU_features_out_of_sample_val
                mask = self.mask_out_of_sample_val
                len_SKUs = len(self.out_of_sample_val_SKUs)
            elif self.return_SKU_type == "out_of_sample_test":
                demand = self.demand_out_of_sample_test
                demand_lag = self.demand_lag_out_of_sample_test
                SKU_features = self.SKU_features_out_of_sample_test
                time_SKU_features = self.time_SKU_features_out_of_sample_test
                mask = self.mask_out_of_sample_test
                len_SKUs = len(self.out_of_sample_test_SKUs)
        
        else:
            demand = self.demand
            demand_lag = self.demand_lag
            SKU_features = self.SKU_features
            time_SKU_features = self.time_SKU_features
            mask = self.mask
            len_SKUs = len(self.train_SKUs)
        
        time_features = self.time_features # time features independent of SKU


        lag_window = self.lag_window_params["lag_window"]
        include_y = self.lag_window_params["include_y"]

        idx_time, idx_skus = self.get_time_SKU_idx(idx)
        num_skus = len(idx_skus)

        demand = demand[idx_time, idx_skus]
        
        item = np.empty((1,lag_window+1, self.num_features, num_skus))

        if include_y:
            assert idx_time-1-lag_window >= 0
            lag_demand = demand_lag[idx_time-1-lag_window:idx_time, idx_skus] # need to use t-1 to get the lag  

        if self.provide_additional_target:
            additional_target = demand_lag[idx_time-lag_window:idx_time+1, idx_skus] # provide target without lag
            additional_target[-1, :] = 0 # The transformer cannot see the last target --> this is to be predicted

        if self.SKU_features is not None:
            SKU_features = SKU_features[idx_skus].transpose()
            len_SKU_features = len(SKU_features)
            SKU_features = np.expand_dims(SKU_features, axis=0)
            SKU_features = np.repeat(SKU_features, repeats=lag_window+1, axis=0) 
        else:
            len_SKU_features = 0

        time_features = time_features[idx_time-lag_window:idx_time+1]
        len_time_features = time_features.shape[1]
        time_features = np.expand_dims(time_features, axis=-1)
        time_features = np.repeat(time_features, num_skus, axis=-1)

        num_time_SKU_features_without_lag_demand = self.num_time_SKU_features-include_y-self.include_non_available-self.provide_additional_target

        for i, idx_sku in enumerate(idx_skus):

            SKU_indices = [len_SKUs*i+idx_sku for i in range(num_time_SKU_features_without_lag_demand)]
            time_SKU_features_sku = time_SKU_features[idx_time-lag_window:idx_time+1, SKU_indices]
            item[:,:,(len_SKU_features+len_time_features):(len_SKU_features+len_time_features+num_time_SKU_features_without_lag_demand),i] = np.expand_dims(time_SKU_features_sku, axis=0)
        
        item[:,:,:len_SKU_features,:] =  np.expand_dims(SKU_features, axis=0)
        
        item[:,:,len_SKU_features:(len_SKU_features+len_time_features),:] = np.expand_dims(time_features, axis=0)

        extra_info = sum([self.include_non_available, include_y, self.provide_additional_target])
        additional_info = np.empty((1,lag_window+1, extra_info, num_skus))

        current_index = 0

        if self.include_non_available:

            additional_info[:, :, current_index, :] = np.expand_dims(mask[idx_time-lag_window:idx_time+1, idx_skus], axis=0)
            current_index += 1

        if include_y:
        
            additional_info[:, :, current_index, :] = np.expand_dims(lag_demand, axis=0)
            current_index += 1
    
        if self.provide_additional_target:
            
            additional_info[:, :, current_index, :] = np.expand_dims(additional_target, axis=0)

        item[:,:,-extra_info:,:] = additional_info

        if self.dataset_type == "train":
            if self.permutate_inputs:
                start_index_to_permutate = len_SKU_features
                end_index_to_permutate = item.shape[2]
                if self.provide_additional_target:
                    end_index_to_permutate -= 1 # target shall always be at the end
                indices_for_permutation = np.arange(start_index_to_permutate, end_index_to_permutate)
                print("original order:", indices_for_permutation)
                indices = np.random.permutation(indices_for_permutation)
                item[:,:,(start_index_to_permutate):end_index_to_permutate,:] = item[:,:,(indices),:]
                print("new order:", indices)

        if self.meta_learn_units:
            if self.dataset_type == "train":
                if item.shape[-1] == 1:
                    item = item.squeeze(-1)
                else:
                    raise ValueError('SKU as batch, but item has more than one SKU dimension')
        else:
            if item.shape[-1] != 1:
                raise ValueError('Num_units dimension must be 1 if not meta-learning')
            else:
                item = item.squeeze(-1)

        if item.shape[0] != 1:
            raise ValueError('Batch dimension must be 1')
        item = item.squeeze(0) # remove batch dimension
 
        return item, demand

    def __len__(self):
        return len(self.demand)
    
    @property
    def X_shape(self):

        if self.meta_learn_units:
            return (len(self.time_features), self.lag_window_params["lag_window"]+1, self.num_features)
        else:
            return (len(self.time_features), self.lag_window_params["lag_window"]+1, self.num_features) # check if there will be a difference.
    
    @property
    def Y_shape(self):
        return (len(self.time_features), self.num_units) # using num of val_test SKUs

    @property
    def len_train(self):
        if self.meta_learn_units:
            return len(self.sku_time_index) # sku_time_index contains only timesteps that are in the training set and skus in the training set.
        else:
            return self.len_train_time

    @property
    def len_val(self):
        if self.val_index_start is None:
            raise ValueError('no validation set defined')
        return self.test_index_start-self.val_index_start # validating and testing is always along the time demension (units are a separate dimension)

    @property
    def len_test(self):
        if self.test_index_start is None:
            raise ValueError('no test set defined')
        return len(self.demand)-self.test_index_start # validating and testing is always along the time demension (units are a separate dimension)

    def get_all_X(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        logging.info("Retrieving all X data")

        """
        Returns the entire features dataset.
        Return either the train, val, test, or all data.
        """

        print("mode of dataloader when getting item:", self.dataset_type)

        if dataset_type == 'train':

            X = np.empty((self.len_train, self.lag_window_params["lag_window"]+1, self.num_features))

            for i in range(self.len_train):
                X[i], _ = self[i]

        elif dataset_type == 'val':
            raise NotImplementedError('Not implemented yet')
        elif dataset_type == 'test':
            raise NotImplementedError('Not implemented yet')
        elif dataset_type == 'all':
            raise NotImplementedError('Not implemented yet')
        else:
            raise ValueError('dataset_type not recognized')

        return X

    def get_all_Y(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        """
        Returns the entire target dataset.
        Return either the train, val, test, or all data.
        """

        print("mode of dataloader when getting item:", self.dataset_type)
        if dataset_type == 'train':
            return self.demand[self.train_index_start:self.val_index_start, self.train_SKUs_indices]
        elif dataset_type == 'val':
            raise NotImplementedError('Not implemented yet')
        elif dataset_type == 'test':
            raise NotImplementedError('Not implemented yet')
        elif dataset_type == 'all':
            raise NotImplementedError('Not implemented yet')
        else:
            raise ValueError('dataset_type not recognized')

    @staticmethod
    def is_one_hot(column):
        return set(column.unique()) <= {0, 1}

    @staticmethod
    def is_one_hot_across_skus(feature_df):
        """
        Check if the set of unique values in a feature across all SKU_ids is {0, 1}.
        feature_df: DataFrame slice for a specific feature with SKU_ids as columns.
        """
        flattened_values = feature_df.values.flatten()
        
        # Check if the unique values in this flattened array are exactly {0, 1}
        unique_values = set(flattened_values)
        
        return unique_values <= {0, 1}

    @staticmethod
    def save_indices(df):
        """
        Saves the row and column indices of a DataFrame.
        """
        return {
            'rows': df.index,
            'columns': df.columns
        }
    
    def set_return_sku(
        self,
        sku_type: str = "in_sample" # if train, then the dataloader returns in-sample SKUs, other option "out_of_sample_val" or "out_of_sample_test"
        ):
       
        """
        Set the SKU type to be returned in the __getitem__ method.
        """

        self.return_SKU_type = sku_type



"""Base class for dataloaders"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/10_dataloaders/10_base_dataloader.ipynb.

# %% auto 0
__all__ = ['BaseDataLoader']

# %% ../../nbs/10_dataloaders/10_base_dataloader.ipynb 3
import numpy as np
from abc import ABC, abstractmethod
from typing import Union

# %% ../../nbs/10_dataloaders/10_base_dataloader.ipynb 4
class BaseDataLoader(ABC):
   
    """
    Base class for data loaders.
    The idea of the data loader is to provide all external information to the environment
    (including lagged data, demand etc.). Internal data influenced by past decisions (like
    inventory levels) is to be added from within the environment
    """

    def __init__(self):
        self.dataset_type = "train"

    @abstractmethod
    def __len__(self):
        '''
        Returns the length of the dataset. For dataloaders based on distributions, this 
        should return an error that the length is not defined, otherwise it should return
        the number of samples in the dataset.
        '''
        pass

    @abstractmethod   
    def __getitem__(self, idx):

        """
        Returns always a tuple of X and Y data. If no X data is available, return None.
        """
        pass

    @property
    @abstractmethod
    def X_shape(self):
        """
        Returns the shape of the X data.
        It should follow the format (n_samples, n_features). If the data has a time dimension with
        a fixed length, the shape should be (n_samples, n_time_steps, n_features). If the data is 
        generated from a distribtition, n_samples should be set to 1.
        """
        pass

    @property
    @abstractmethod
    def Y_shape(self):
        """
        Returns the shape of the Y data.
        It should follow the format (n_samples, n_SKUs). If the variable of interst is only a single
        SKU, the shape should be (n_samples, 1). If the data is 
        generated from a distribtition, n_samples should be set to 1.
        """
        pass

    @abstractmethod   
    def get_all_X(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        """
        Returns the entire features dataset. If no X data is available, return None.
        Return either the train, val, test, or all data.
        """
        pass    

    @abstractmethod   
    def get_all_Y(self,
                dataset_type: str = 'train' # can be 'train', 'val', 'test', 'all'
                ): 

        """
        Returns the entire target dataset. If no Y data is available, return None.
        Return either the train, val, test, or all data.
        """
        pass  

    @property
    @abstractmethod   
    def len_train(self):

        """
        Returns the length of the training set. For dataloaders based on distributions, this
        should return an error that the length is not defined, otherwise it should return
        the number of samples in the training set.
        """

        pass

    @property
    @abstractmethod 
    def len_val(self):

        """

        Returns the length of the validation set. For dataloaders based on distributions, this
        should return an error that the length is not defined, otherwise it should return
        the number of samples in the validation set.

        If no valiation set is defined, raise an error.
        """

        pass
    
    @property
    @abstractmethod 
    def len_test(self):
        
        """

        Returns the length of the test set. For dataloaders based on distributions, this
        should return an error that the length is not defined, otherwise it should return
        the number of samples in the test set.

        If no test set is defined, raise an error.
        """

        pass

        
    def train(self):

        """
        Set the internal state of the dataloader to train
        """

        self.dataset_type = "train"

    def val(self):

        """
        Set the internal state of the dataloader to validation
        """

        if self.val_index_start is None:
            raise ValueError('no validation set defined')
        else:
            self.dataset_type = "val"

    def test(self):

        """
        Set the internal state of the dataloader to test
        """

        if self.test_index_start is None:
            raise ValueError('no test set defined')
        else:
            self.dataset_type = "test"




"""Functions that help with tracking of experiments"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils/30_tracking.ipynb.

# %% auto 0
__all__ = ['get_git_hash', 'get_library_version']

# %% ../nbs/00_utils/30_tracking.ipynb 3
from typing import Union, List, Tuple, Literal
import pkg_resources
import subprocess
import wandb
import logging

# %% ../nbs/00_utils/30_tracking.ipynb 4
def get_git_hash(
    directory: str, # the directory where the git repository is located
    tracking: bool = False, # whether to directly track the git revision hash
    tracking_tool: Literal['wandb'] = 'wandb' # Currently only wandb is supported
) -> str:
    
    """ Get the git hash and optionally track """

    hash = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=directory).decode('ascii').strip()

    if tracking:
        if tracking_tool == 'wandb':
            wandb.config.update({'git_hash': hash})
        else:
            raise ValueError(f"Tracking tool {tracking_tool} is not supported")

    return hash

# %% ../nbs/00_utils/30_tracking.ipynb 6
def get_library_version(
    library_name: str,
    tracking: bool = False, # Whether to directly track the library version
    tracking_tool: Literal['wandb'] = 'wandb' # Currently only wandb is supported
    ) -> str:

    """ Get the version of a specified library and optionally track it """
    
    try:
        version = pkg_resources.get_distribution(library_name).version
    except pkg_resources.DistributionNotFound:
        logging.warning(f"Library '{library_name}' not found")
        return "Not installed or not found"

    if tracking:
        if tracking_tool == 'wandb':
            wandb.config.update({f'{library_name}_version': version})
        else:
            raise ValueError(f"Tracking tool {tracking_tool} is not supported")

    return version



"""Class to load datasets available in GitHub releases of this repository."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/80_datasets/default_datasets.ipynb.

# %% auto 0
__all__ = ['get_all_release_tags', 'get_release_tag', 'get_dataset_url', 'get_asset_url', 'download_file_from_github',
           'unzip_file', 'load_data_from_directory', 'DatasetLoader']

# %% ../../nbs/80_datasets/default_datasets.ipynb 3
import numpy as np
import logging
import requests
import os
import re
import pandas as pd
import zipfile


# %% ../../nbs/80_datasets/default_datasets.ipynb 6
def get_all_release_tags(token=None):
    url = "https://api.github.com/repos/opimwue/ddopai/releases"
    headers = {'Authorization': f'Bearer {token}'} if token else {}
    response = requests.get(url, headers=headers)
    
    if response.status_code == 200:
        releases = response.json()
        tags = [release['tag_name'] for release in releases]
        return tags
    else:
        raise ValueError(f"Failed to fetch releases: {response.status_code} with message: {response.text}")

def get_release_tag(dataset_type, version, token=None):
    release_tags = get_all_release_tags(token)
    release_tags_filtered = [tag for tag in release_tags if dataset_type in tag]

    if version == "latest":
        release_tags_filtered.sort(key=lambda x: [int(num) if num.isdigit() else num for num in re.findall(r'\d+|\D+', x.split('_v')[-1])])
        release_tag = release_tags_filtered[-1]
    else:
        release_tag = f"{dataset_type}_{version}"
    
    logging.debug(f"Filtered release tags: {release_tags_filtered}")
    return release_tag

def get_dataset_url(dataset_type, dataset_number, release_tag, token=None):
    api_url = f"https://api.github.com/repos/opimwue/ddopai/releases/tags/{release_tag}"
    headers = {'Authorization': f'Bearer {token}'} if token else {}
    response = requests.get(api_url, headers=headers)

    if response.status_code == 200:
        release_info = response.json()
        assets = release_info.get("assets", [])

        assets = [asset for asset in assets if f"{dataset_type}_dataset_{dataset_number}_" in asset['name']]

        for asset in assets:
            logging.debug(f"Found dataset: {asset['name']}")

        if len(assets) == 0:
            raise ValueError(f"Dataset {dataset_type}_dataset_{dataset_number} not found in release {release_tag}")
        elif len(assets) > 1:
            raise ValueError(f"Multiple datasets found for {dataset_type}_dataset_{dataset_number} in release {release_tag}")
        else:
            asset = assets[0]
            return asset['browser_download_url']
    else:
        raise ValueError(f"Failed to fetch release information: {response.status_code} with message: {response.text}")

def get_asset_url(dataset_type, dataset_number, version="latest", token=None):
    release_tag = get_release_tag(dataset_type, version, token)
    asset_url = get_dataset_url(dataset_type, dataset_number, release_tag, token)
    return asset_url

def download_file_from_github(url, output_path, token=None):
    headers = {'Authorization': f'Bearer {token}'} if token else {}
    response = requests.get(url, headers=headers, stream=True)
    if response.status_code == 200:
        with open(output_path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    file.write(chunk)
        logging.debug(f"File downloaded successfully: {output_path}")
    else:
        logging.error(f"Failed to download file: {response.status_code}")

def unzip_file(zip_file_path, output_dir, delete_zip_file=True):
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        zip_ref.extractall(output_dir)

    if delete_zip_file:
        os.remove(zip_file_path)

def load_data_from_directory(dir):
    data = dict()
    for file in os.listdir(dir):
        if file.endswith(".csv"):
            key = os.path.splitext(file)[0]
            data[key] = pd.read_csv(os.path.join(dir, file))
        elif file.endswith(".pkl"):
            key = os.path.splitext(file)[0]
            data[key] = pd.read_pickle(os.path.join(dir, file))
        elif file.endswith(".npy"):
            key = os.path.splitext(file)[0]
            data[key] = np.load(os.path.join(dir, file))
        else:
            raise ValueError(f"File {file} is not a valid file type (csv, pkl, or npy)")
    
    return data


# %% ../../nbs/80_datasets/default_datasets.ipynb 8
class DatasetLoader():

    """
    Class to load datasets from the GitHub repository.
    """

    dataset_types_univariate = [
    ]

    dataset_types_multivariate = [
        "arma_10_10",
        "arma_2_2",
        "ar_1",
    ]
    
    def __init__(self):
        pass
    
    def show_dataset_types(self,
            show_num_datasets_per_type=False # Whether to show the number of datasets per type
            ):

        """ Show an overview of all dataset types available in the repository."""

        if show_num_datasets_per_type:
            raise NotImplementedError("show_num_datasets_per_type is not implemented yet.")
        else:
            print("Univariate datasets:")
            for dataset in DatasetLoader.dataset_types_univariate:
                print(dataset)
            
            print("\nMultivariate datasets:")
            for dataset in DatasetLoader.dataset_types_multivariate:
                print(dataset)
        
    def load_dataset(self,
        dataset_type: str,
        dataset_number: int,
        overwrite: bool = False, # Whether to overwrite the dataset if it already exists
        version: str = "latest", # Which version of the dataset to load, "latest" or a specific version,
        token: str = None # GitHub token to enable more requests (otherwise limited to 60 requests per hour)
    ):

        """ Load a dataset from the GitHub repository."""

        if dataset_type not in self.dataset_types_univariate and dataset_type not in self.dataset_types_multivariate:
            raise ValueError(f"Dataset type {dataset_type} is not valid. Use the function show_dataset_types() to see valid dataset types.")

        asset_url = get_asset_url(dataset_type, dataset_number, version=version, token=token)


        # check if folder "data" exists, if not create it
        if not os.path.exists("data"):
            os.makedirs("data")

        output_file_path = f"data/{dataset_type}_dataset_{dataset_number}"

        download = False
        # check if the dataset has already been downloaded
        if os.path.exists(output_file_path):
            logging.warning(f"Dataset {dataset_type}_dataset_{dataset_number} has already been downloaded.")
            if overwrite:
                logging.warning("Overwriting dataset.")
                download = True
            else:
                logging.warning("Keeping existing dataset.")
        else:
            download = True

        if download:
            download_file_from_github(asset_url, output_file_path+".zip", token=token)
            unzip_file(output_file_path+".zip", output_file_path)

        data = load_data_from_directory(output_file_path)

        return data



"""Dataloaders that return data by sampling from some pre-defined distributions."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/10_dataloaders/11_distribution_loaders.ipynb.

# %% auto 0
__all__ = ['BaseDistributionDataLoader', 'NormalDistributionDataLoader']

# %% ../../nbs/10_dataloaders/11_distribution_loaders.ipynb 3
import numpy as np
from abc import ABC, abstractmethod
from typing import Union

from .base import BaseDataLoader

# %% ../../nbs/10_dataloaders/11_distribution_loaders.ipynb 4
class BaseDistributionDataLoader(BaseDataLoader):

    is_distribution = True

    def __init__(self):
        super().__init__()

# %% ../../nbs/10_dataloaders/11_distribution_loaders.ipynb 5
class NormalDistributionDataLoader(BaseDistributionDataLoader):

    """
    A dataloader that generates a dataset of normally distributed values.
    """
    
    def __init__(self,
        mean: float,
        std: float,
        num_units: int,
        truncated_low: int = 0,
        truncated_high: int = None #
    ):
        self.num_units = num_units
        self.mean = mean
        self.std = std
        self.truncated_low = truncated_low
        self.truncated_high = truncated_high

        self.val_index_start = 0 # No special validation set, necessary such that dataloader.val() does not throw an error
        self.test_index_start = 0  # No special test set, necessary such that dataloader.test() does not throw an error
        
        super().__init__()
    
    def __getitem__(self, idx):

        """
    
        Samples a datapoint from the distribution. As the distribution is generated on the fly, the index is not used.
        """

        Y = np.random.normal(self.mean, self.std, self.num_units)

        if self.truncated_low is not None:
            Y = np.maximum(Y, self.truncated_low)
        if self.truncated_high is not None:
            Y = np.minimum(Y, self.truncated_high)

        return None, Y

    def __len__(self):
        """
        Returns the length of the distribution. As the distribution is generated on the fly, the length is not defined.
        """
        raise ValueError('Length of a distribution is not defined')

    @property
    def X_shape(self):
        return None

    @property
    def Y_shape(self):
        return (1, self.num_units)
 
    def get_all_X(self):

        """
        Returns the entire features dataset. If no X data is available, return None.
        """
        return None

    def get_all_Y(self):

        """
        Returns the entire target dataset. If no Y data is available, return None.
        """
        return None

    @property
    def len_train(self):
        return np.inf

    @property
    def len_val(self):
        return np.inf

    @property
    def len_test(self):
        return np.inf



"""Functions to run experiments more efficiently. The usage of these functions is optional and they are only compatible with agents defined in this package. Using agents from other packages such as Stable Baselines or RLlib may require using their own experiment functions."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/30_experiment_functions/10_experiment_functions.ipynb.

# %% auto 0
__all__ = ['EarlyStoppingHandler', 'calculate_score', 'log_info', 'update_best', 'save_agent', 'test_agent', 'run_test_episode',
           'run_experiment']

# %% ../nbs/30_experiment_functions/10_experiment_functions.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, List, Tuple, Dict, Literal
import logging
from datetime import datetime  
import numpy as np
import sys
import wandb

from .envs.base import BaseEnvironment
from .agents.base import BaseAgent

import importlib

from tqdm import tqdm, trange

# Think about how to handle mushroom integration.
from mushroom_rl.core import Core

# %% ../nbs/30_experiment_functions/10_experiment_functions.ipynb 4
class EarlyStoppingHandler():

    '''
    Class to handle early stopping during experiments. The EarlyStoppingHandler handler calculates the average
    score over the last "patience" epochs and compares it to the average score over the previous "patience" epochs.
    Note that one epoch we define here as time in between evaluating on a validation set, for supervised learning
    typically one epoch is one pass through the training data. For reinforcement learning, in between each evaluation
    epoch there may be less than one, one, or many episodes played in the training environment.

    '''
    def __init__(
        self,
        patience: int = 50, # Number of epochs to evaluate for stopping
        warmup: int = 100, # How many initial epochs to wait before evaluating
        criteria: str = "J",  # Whether to use discounted rewards J or total rewards R as criteria
        direction: str = "max"  # Whether reward shall be maximized or minimized
    ):

        self.history = list()
        self.patience = patience
        if warmup is None or warmup < patience * 2:
            warmup = patience * 2
        self.warmup = warmup
        self.criteria = criteria
        self.direction = direction

    def add_result(self,
                    J: float, # Return (discounted rewards) of the last epoch
                    R: float, # Total rewards of the last epoch
                    ) -> bool:

        """
        Add the result of the last epoch to the history and check if the experiment should be stopped.

        """
        if self.criteria == "J":
            self.history.append(J)
        elif self.criteria == "R":
            self.history.append(R)
        else:
            raise ValueError("Criteria must be J or R")
        
        if len(self.history) >= self.warmup:
            if self.direction == "max":
                if sum(self.history[-self.patience*2:-self.patience]) >= sum(self.history[-self.patience:]):
                    return True
                else:
                    return False
            elif self.direction == "min":
                if sum(self.history[-self.patience*2:-self.patience]) <= sum(self.history[-self.patience:]):
                    return True
                else:
                    return False
            else:
                raise ValueError("Direction must be max or min")

# %% ../nbs/30_experiment_functions/10_experiment_functions.ipynb 8
def calculate_score(
                    dataset: List,
                    env: BaseEnvironment, # Any environment inheriting from BaseEnvironment
                    ) -> Tuple[float, float]:

    """

    Calculate the total rewards R and the discounted rewards J of a dataset.

    """

    R = sum([row[0][2] for row in dataset])
    gamma = env.mdp_info.gamma
    J = sum([gamma**(t) * row[0][2] for t, row in enumerate(dataset)]) # Note: t starts at 1 so the first reward is already discounted

    return R, J

def log_info(R: float,
                J: float,
                n_epochs: int,
                tracking: Literal["wandb"], # only wandb implemented so far
                mode: Literal["train", "val", "test"]
                ):
    
    '''
    Logs the same R, J information repeatedly for n_epoochs.
    E.g., to draw a straight line in wandb for algorithmes
    such as XGB, RF, etc. that can be comparared to the learning
    curves of supervised or reinforcement learning algorithms.
    '''

    if tracking == "wandb":
        for epoch in range(n_epochs):
            wandb.log({f"{mode}/R": R, f"{mode}/J": J})
    else:
        pass

def update_best(R: float, J: float, best_R: float, best_J: float): # 
    
    """

    Update the best total rewards R and the best discounted rewards J.

    """

    if R > best_R:
        best_R = R
    if J > best_J:
        best_J = J

    return best_R, best_J

def save_agent(agent: BaseAgent, # Any agent inheriting from BaseAgent
                experiment_dir: str, # Directory to save the agent, 
                save_best: bool,
                R: float,
                J: float,
                best_R: float,
                best_J: float,
                criteria: str = "J",
                force_save = False,
                ):

    """
    Save the agent if it has improved either R or J, depending on the criteria argument,
    vs. the previous epochs

    """

    if save_best:
        if criteria == "R":
            if R == best_R:
                save_dir = f"{experiment_dir}/saved_models/best"
                agent.save(save_dir)
            elif force_save:
                save_dir = f"{experiment_dir}/saved_models/best"
                agent.save(save_dir)
        elif criteria == "J":
            if J == best_J:
                save_dir = f"{experiment_dir}/saved_models/best"
                agent.save(save_dir)
            elif force_save:
                save_dir = f"{experiment_dir}/saved_models/best"
                agent.save(save_dir)

# %% ../nbs/30_experiment_functions/10_experiment_functions.ipynb 10
def test_agent(agent: BaseAgent,
            env: BaseEnvironment,
            return_dataset = False,
            save_features = False,
            tracking = None, # other: "wandb",
            eval_step_info = False,
):

    """
    Tests the agent on the environment for a single episode
    """
    
    # TODO make it possible to save dataset via tracking tool

    # Run the test episode
    dataset = run_test_episode(env, agent, eval_step_info, save_features = save_features)

    # Calculate the score
    R, J = calculate_score(dataset, env)

    if tracking == "wandb":
        mode = env.mode
        wandb.log({f"{mode}/R": R, f"{mode}/J": J})

    if return_dataset:
        return R, J, dataset
    else:
        return R, J

def run_test_episode(   env: BaseEnvironment, # Any environment inheriting from BaseEnvironment
                        agent: BaseAgent, # Any agent inheriting from BaseAgent
                        eval_step_info: bool = False, # Print step info during evaluation
                        save_features: bool = False, # Save features (observation) of the dataset. Can be turned off since they sometimes become very large with many lag information

                ):

    """
    Runs an episode to test the agent's performance.
    It assumes, that agent and environment are initialized, in test/val mode
    and have done reset.
    """

    # Get initial observation
    obs = env.reset()

    dataset = []
    
    finished = False
    step = 0

    horizon = env.mdp_info.horizon
    
    while not finished:
        
        # Sample action from agent
        action = agent.draw_action(obs)

        # Take a step in the environment

        next_obs, reward, terminated, truncated, info = env.step(action)
        
        logging.debug("##### STEP: %d #####", env.index)
        logging.debug("reward: %s", reward)
        logging.debug("info: %s", info)
        logging.debug("next observation: %s", obs)
        logging.debug("truncated: %s", truncated)

        if save_features:
            sample = (obs, action, reward, next_obs, terminated, truncated) # unlike mushroom do not include policy_state
        else:
            sample = (None, action, reward, None, terminated, truncated)

        obs = next_obs
        
        dataset.append((sample, info))

        finished = terminated or truncated

        if eval_step_info:
            step += 1
            sys.stdout.write(f"\rStep {step}")
            sys.stdout.flush()

    if eval_step_info:
        print()

    return dataset

def run_experiment( agent: BaseAgent,
                    env: BaseEnvironment,

                    n_epochs: int,
                    n_steps: int = None, # Number of steps to interact with the environment per epoch. Will be ignored for direct_fit and epchos_fit agents

                    early_stopping_handler: Union[EarlyStoppingHandler, None] = None,
                    save_best: bool = True,
                    performance_criterion: str = "J", # other: "R"

                    tracking: Union[str, None]  = None, # other: "wandb"

                    results_dir: str = "results",

                    run_id: Union[str, None] = None,

                    print_freq: int = 10,

                    eval_step_info = False,

                    return_score = False,
                ):

    """
    Run an experiment with the given agent and environment for n_epochs. It automaticall dedects if the train mode
    of the agent is direct, epochs_fit or env_interaction and runs the experiment accordingly.

    """

    if return_score:
        R_list = []
        J_list = []

    # use start_time as id if no run_id is given
    if run_id is None:
        run_id = datetime.now().strftime("%Y%m%d_%H%M%S_%f")

    experiment_dir = f"{results_dir}/{run_id}"

    print(f"Experiment directory: {experiment_dir}")

    logging.info("Starting experiment")

    env.reset()

    # initial evaluation
    env.val()
    agent.eval()
    R, J = test_agent(agent, env, tracking = tracking)

    env.train()
    agent.train()

    logging.info(f"Initial evaluation: R={R}, J={J}")

    best_J = J 
    best_R = R

    if agent.train_mode == "direct_fit":
        
        logging.info("Starting training with direct fit")
        agent.fit(X=env.dataloader.get_all_X("train"), Y=env.dataloader.get_all_Y("train"))
        logging.info("Finished training with direct fit")

        env.val()
        agent.eval()

        R, J = test_agent(agent, env, tracking = tracking, eval_step_info=eval_step_info)
        best_R, best_J = update_best(R, J, best_R, best_J)

        logging.info(f"Evaluation after training: R={R}, J={J}")

        save_agent(agent, experiment_dir, save_best, R, J, best_R, best_J, performance_criterion, force_save = True) # save even if not best

        log_info(R, J, n_epochs-1, tracking, "val")

        if return_score:
            R_list.append(R)
            J_list.append(J)

    elif agent.train_mode == "epochs_fit":

        # save initial agent
        save_dir = f"{experiment_dir}/saved_models/best"
        agent.save(save_dir)
        
        logging.info("Starting training with epochs fit")
        for epoch in trange(n_epochs):
            
            agent.fit_epoch() # Access to dataloader provided to the agent at initialization

            env.val()
            agent.eval()

            R, J = test_agent(agent, env, tracking = tracking, eval_step_info=eval_step_info)

            if return_score:
                R_list.append(R)
                J_list.append(J)
            
            if ((epoch+1) % print_freq) == 0:
                logging.info(f"Epoch {epoch+1}: R={R}, J={J}")
            
            best_R, best_J = update_best(R, J, best_R, best_J)
            save_agent(agent, experiment_dir, save_best, R, J, best_R, best_J, performance_criterion)
            
            if early_stopping_handler is not None:
                stop = early_stopping_handler.add_result(J, R)
            else:
                stop = False

            if stop:
                log_info(R, J, n_epochs-epoch-1, tracking, "val")
                logging.info(f"Early stopping after {epoch+1} epochs")
                break
        
            env.train()
            agent.train()

        logging.info("Finished training with epochs fit")

    elif agent.train_mode == "env_interaction":

        # save initial agent
        save_dir = f"{experiment_dir}/saved_models/best"
        agent.save(save_dir)

        logging.info("Starting training with env_interaction")

        core = Core(agent, env)

        agent.train()
        env.train()

        if hasattr(agent, "warmup_training_steps"):
            warmup_training = True
            warmup_training_steps = agent.warmup_training_steps
        else:
            warmup_training = False
        
        if hasattr(agent, "n_steps_per_fit"):
            n_steps_per_fit = agent.n_steps_per_fit
        else:
            n_steps_per_fit = 1

        if warmup_training:
            env.set_return_truncation(False) # For mushroom Core to work, the step function should not return the truncation flag
            core.learn(n_steps=warmup_training_steps, n_steps_per_fit=warmup_training_steps, quiet=True)
        
        for epoch in trange(n_epochs):

            env.set_return_truncation(False) # For mushroom Core to work, the step function should not return the truncation flag
            agent.train()
            core.learn(n_steps=n_steps, n_steps_per_fit=n_steps_per_fit, quiet=True)
            env.set_return_truncation(True) # Set back to standard gynmasium behavior

            env.val()
            agent.eval()

            R, J = test_agent(agent, env, tracking = tracking, eval_step_info=eval_step_info)

            if return_score:
                R_list.append(R)
                J_list.append(J)

            if ((epoch+1) % print_freq) == 0:
                logging.info(f"Epoch {epoch+1}: R={R}, J={J}")
            
            best_R, best_J = update_best(R, J, best_R, best_J)
            save_agent(agent, experiment_dir, save_best, R, J, best_R, best_J, performance_criterion)

            if early_stopping_handler is not None:
                stop = early_stopping_handler.add_result(J, R)
            else:
                stop = False

            if stop:
                log_info(R, J, n_epochs-epoch-1, tracking, "val")
                logging.info(f"Early stopping after {epoch+1} epochs")
                break
        
            env.train()
            agent.train()

    else:
        raise ValueError("Unknown train mode")

    if return_score:
        return R_list, J_list

    logging.info(f"Evaluation after training: R={R}, J={J}")



"""Loss functions that are implemented in PyTorch"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/00_utils/20_torch_loss_functions.ipynb.

# %% auto 0
__all__ = ['quantile_loss', 'TorchQuantileLoss', 'pinball_loss', 'TorchPinballLoss']

# %% ../../nbs/00_utils/20_torch_loss_functions.ipynb 3
from typing import Union, Optional, Tuple

import numpy as np
from ..utils import Parameter

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.loss import _Loss

import warnings

# %% ../../nbs/00_utils/20_torch_loss_functions.ipynb 4
def quantile_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    quantile: torch.Tensor,
    reduction: str = 'mean',
) -> torch.Tensor:

    if not (target.size() == input.size()):
        warnings.warn(
            f"Using a target size ({target.size()}) that is different to the input size ({input.size()}). "
            "This will likely lead to incorrect results due to broadcasting. "
            "Please ensure they have the same size.",
            stacklevel=2,
        )

    expanded_input, expanded_target = torch.broadcast_tensors(input, target)

    # print(expanded_input.size(), expanded_target.size(), quantile.size())
    # print(quantile)

    loss = torch.max((expanded_target - expanded_input) * quantile, (expanded_input - expanded_target) * (1 - quantile))

    # print(losks.size())

    if reduction == 'mean':
        return loss.mean()
    elif reduction == 'sum':
        return loss.sum()
    else:
        raise ValueError(f"reduction={reduction} is not valid")

    return loss

# %% ../../nbs/00_utils/20_torch_loss_functions.ipynb 5
class TorchQuantileLoss(_Loss):

    """
    Implmentation of the quantile loss in Pytorch.
    Unlike the Numpy-based implementation ```quantile_loss``` in the loss_functions module, this implementation
    this implementation reduces the results to a scalar value using the specified reduction method. This class is 
    used to train Pytorch models using the quantile loss.
    """

    __constants__ = ['reduction']
    def __init__(self,reduction: str = 'mean') -> None: #
        super().__init__(reduction=reduction)

    def forward(self, input: torch.Tensor, target: torch.Tensor, quantile: Parameter | np.ndarray) -> torch.Tensor: #

        """
        Forward pass of the quantile loss function.

        """

        quantile = self.convert_quantile(quantile, input_dtype=input.dtype, device=input.device, target_shape = target.shape)
            
        if not (target.shape == input.shape == quantile.shape):
            warnings.warn(
                f"Mismatch in dimensions: target dimension ({target.shape}), input dimension ({input.shape}), "
                f"and quantile dimension ({quantile.shape}) must be the same. "
                "This will likely lead to incorrect results due to broadcasting. "
                "Please ensure they have the same size.",
                stacklevel=2,
            )

        return quantile_loss(input, target, quantile, reduction=self.reduction)
    
    def convert_quantile(self, quantile: Parameter | np.ndarray, input_dtype: torch.dtype = torch.float32, device: torch.device = torch.device('cpu'), target_shape: Tuple = None) -> torch.Tensor:
        
        if isinstance(quantile, Parameter):
            quantile =  quantile.get_value()
            
        if isinstance(quantile, np.ndarray):
            quantile = torch.tensor(quantile, dtype=input_dtype, device=device)
        elif isinstance(quantile, torch.Tensor):
            # ensure dtype and device are the same as the input tensor
            quantile = quantile.to(dtype=input_dtype, device=device)
        else:
            raise ValueError(f"quantile must be of type Parameter, np.ndarray, or torch.Tensor, but got {type(quantile)}")

        # check if quantile is of size 1:
        if quantile.size() == (1,):
            quantile = quantile.expand(target_shape)
        elif quantile.size() != target_shape:
            raise ValueError(f"quantile must be of size 1 or the same size as the target tensor, but got {quantile.size()} and {target_shape}")

        return quantile


# %% ../../nbs/00_utils/20_torch_loss_functions.ipynb 9
def pinball_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    underage: torch.Tensor,
    overage: torch.Tensor,
    reduction: str = 'mean',
) -> torch.Tensor:

    if not (target.size() == input.size()):
        warnings.warn(
            f"Using a target size ({target.size()}) that is different to the input size ({input.size()}). "
            "This will likely lead to incorrect results due to broadcasting. "
            "Please ensure they have the same size.",
            stacklevel=2,
        )

    expanded_input, expanded_target = torch.broadcast_tensors(input, target)

    # loss = torch.max((expanded_target - expanded_input) * quantile, (expanded_input - expanded_target) * (1 - quantile))

    loss = torch.max(expanded_target - expanded_input, torch.tensor(0)) * underage + torch.max(expanded_input - expanded_target, torch.tensor(0)) * overage

    if reduction == 'mean':
        return loss.mean()
    elif reduction == 'sum':
        return loss.sum()
    else:
        raise ValueError(f"reduction={reduction} is not valid")

    return loss

# %% ../../nbs/00_utils/20_torch_loss_functions.ipynb 11
class TorchPinballLoss(_Loss):

    """
    Implmentation of the pinball loss in Pytorch using specific overage and underage cost. For the pinball loss
    based on quantiles directly, use the TorchQuantileLoss class.
    Unlike the Numpy-based implementation ```pinball_loss``` in the loss_functions module, this implementation
    this implementation reduces the results to a scalar value using the specified reduction method. This class is 
    used to train Pytorch models using the pinball loss.
    """

    __constants__ = ['reduction']
    def __init__(self,reduction: str = 'mean') -> None: #
        super().__init__(reduction=reduction)

    def forward(self, input: torch.Tensor, target: torch.Tensor, underage: Parameter | np.ndarray, overage: Parameter | np.ndarray) -> torch.Tensor: #

        """
        Forward pass of the pinball loss function.

        """

        underage = self.convert_cost_param(underage, input_dtype=input.dtype, device=input.device, target_shape = target.shape)
        overage = self.convert_cost_param(overage, input_dtype=input.dtype, device=input.device, target_shape = target.shape)
            
        if not (target.shape == input.shape == underage.shape == overage.shape):
            warnings.warn(
                f"Mismatch in dimensions: target dimension ({target.shape}), input dimension ({input.shape}), "
                f"underage dimension ({underage.shape}), and overage dimension ({overage.shape}) must be the same. "
                "This will likely lead to incorrect results due to broadcasting. "
                "Please ensure they have the same size.",
                stacklevel=2,
            )

        return pinball_loss(input, target, underage=underage, overage=overage, reduction=self.reduction)
    
    def convert_cost_param(self, cost_param: Parameter | np.ndarray, input_dtype: torch.dtype = torch.float32, device: torch.device = torch.device('cpu'), target_shape: Tuple = None) -> torch.Tensor:
        
        if isinstance(cost_param, Parameter):
            cost_param =  cost_param.get_value()

        if isinstance(cost_param, np.ndarray):
            cost_param = torch.tensor(cost_param, dtype=input_dtype, device=device)
        elif isinstance(cost_param, torch.Tensor):
            # ensure dtype and device are the same as the input tensor
            cost_param = cost_param.to(dtype=input_dtype, device=device)
        else:
            raise ValueError(f"quantile must be of type Parameter, np.ndarray, or torch.Tensor, but got {type(cost_param)}")

        # check if quantile is of size 1:
        if cost_param.size() == (1,):
            cost_param = cost_param.expand(target_shape)
        elif cost_param.size() != target_shape:
            raise ValueError(f"quantile must be of size 1 or the same size as the target tensor, but got {cost_param.size()} and {target_shape}")

        return cost_param



"""Models that approximate a function"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/60_approximators/11_approximators.ipynb.

# %% auto 0
__all__ = ['BaseModule', 'LinearModel', 'MLP', 'Transformer', 'LlamaRotaryEmbedding', 'rotate_half', 'apply_rotary_pos_emb',
           'CausalSelfAttention', 'find_multiple', 'MLP_block', 'RMSNorm', 'Block']

# %% ../nbs/60_approximators/11_approximators.ipynb 3
# import logging
# logging_level = logging.DEBUG

from abc import ABC, abstractmethod
from typing import Union, Dict, Literal
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader

import time

# %% ../nbs/60_approximators/11_approximators.ipynb 4
# TODO: Merge with base from RL networks to avoid duplicaiotn of code

class BaseModule(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def select_activation(activation):
        """ Select the activation function based on input string """
        activation = activation.lower()  # Convert input to lowercase for consistency
        if activation == "relu":
            return nn.ReLU
        elif activation == "sigmoid":
            return nn.Sigmoid
        elif activation == "tanh":
            return nn.Tanh
        elif activation == "elu":
            return nn.ELU
        elif activation == "leakyrelu":
            return nn.LeakyReLU
        elif activation == "identity":
            return nn.Identity
        else:
            raise ValueError(f"Activation function {activation} not recognized")

# %% ../nbs/60_approximators/11_approximators.ipynb 5
class LinearModel(BaseModule):
    """Linear regression model"""

    def __init__(self, 
            input_size: int, # number of features
            output_size: int, # number of outputs/actions
            relu_output: bool = False): # whether to apply ReLU activation to the output
        super().__init__()
        self.l1=nn.Linear(input_size, output_size)
        if relu_output:
            self.final_activation = nn.ReLU()
        else:
            self.final_activation = nn.Identity()
            
    def forward(self,x):
        out=self.l1(x)
        out=self.final_activation(out)
        return out

# %% ../nbs/60_approximators/11_approximators.ipynb 6
class MLP(BaseModule):

    """ Multilayer perceptron model """

    def __init__(self,
                    input_size: int, # number of features
                    output_size: int, # number of outputs/actions
                    hidden_layers: list, # list of number of neurons in each hidden layer
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    relu_output: bool = False): # whether to apply ReLU activation to the output
        super().__init__()

        # List of layers
        layers = []

        last_size = input_size
        for num_neurons in hidden_layers:
            layers.append(nn.Linear(last_size, num_neurons))
            layers.append(nn.ReLU())
            if batch_norm:
                layers.append(nn.BatchNorm1d(num_neurons))
            layers.append(nn.Dropout(p=drop_prob))
            last_size = num_neurons

        # Output layer
        layers.append(nn.Linear(last_size, output_size))
        if relu_output:
            layers.append(nn.ReLU()) # output is non-negative
        else:
            layers.append(nn.Identity())

        # Combine layers
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)

# %% ../nbs/60_approximators/11_approximators.ipynb 7
class Transformer(BaseModule):

    """ Multilayer perceptron model """

    def __init__(self,
                    input_size: int, # number of (time steps, features)
                    output_size: int, # number of outputs/actions

                    max_context_length: int = 128, #  maximum context lenght during inference
                    n_layer: int = 3, # number of layers in the transformer
                    n_head: int = 8, # number of heads per layer
                    n_embd_per_head: int = 32, # number of embedding per head
                    rope_scaling: Dict | None = None, # whether to use rope scaling, not implemented yet

                    min_multiple = 256, # minimum multiple for neurons in the MLP block of the transformer
                    gating = True, # Whether to apply the gating mechanism from the original Llama model (used in LagLlama)

                    drop_prob: float = 0.0, # dropout probability
                    final_activation: Literal["relu", "sigmoid", "tanh", "elu", "leakyrelu", "identity"] = "identity" # final activation function
                    ): # whether to apply ReLU activation to the output

        super().__init__()

        block_size = max_context_length
        input_size = input_size[1] # we only consider the number of features

        self.param_proj = nn.Linear(n_embd_per_head * n_head, output_size) # final projection layer for output

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Linear(
                    input_size, n_embd_per_head * n_head # Initial projection from input to embedding space
                ),
                h=nn.ModuleList([Block(n_embd_per_head, n_head, block_size, drop_prob, min_multiple = min_multiple, gating=gating) for _ in range(n_layer)]),
                ln_f=RMSNorm(n_embd_per_head * n_head),
            )
        )

        self.final_activation = self.select_activation(final_activation)()

        # not _init_weights used since we are using the default initialization.

    def forward(    self,
                    x: torch.Tensor,) -> torch.Tensor:

        (B, T, C) = x.size()

        x = self.transformer.wte(
            x
        )

        for block in self.transformer.h:
            x = block(x)

        output = self.param_proj(
            x
        ) 

        output = self.final_activation(output)

        output = output[:, -1, :] # we use the last time dimension as the output
             
        return output

# %% ../nbs/60_approximators/11_approximators.ipynb 8
class LlamaRotaryEmbedding(torch.nn.Module):

    """
    Rotary positional embeddings (RoPE) based on https://arxiv.org/abs/2104.09864
    Code following the implementation in https://github.com/time-series-foundation-models/lag-llama

    """

    def __init__(self, dim, max_position_embeddings=2048, base=10000, device=None):
        super().__init__()

        self.dim = dim
        self.max_position_embeddings = max_position_embeddings
        self.base = base
        inv_freq = 1.0 / (
            self.base ** (torch.arange(0, self.dim, 2).float().to(device) / self.dim)
        )
        self.register_buffer("inv_freq", inv_freq, persistent=False)

        self._set_cos_sin_cache(
            seq_len=max_position_embeddings,
            device=self.inv_freq.device,
            dtype=torch.get_default_dtype(),
        )

    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len
        t = torch.arange(
            self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype
        )

        freqs = torch.einsum("i,j->ij", t, self.inv_freq)
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer(
            "cos_cached", emb.cos()[None, None, :, :].to(dtype), persistent=False
        )
        self.register_buffer(
            "sin_cached", emb.sin()[None, None, :, :].to(dtype), persistent=False
        )

    def forward(self, device, dtype, seq_len=None):
        if seq_len > self.max_seq_len_cached:
            self._set_cos_sin_cache(seq_len=seq_len, device=device, dtype=dtype)

        return (
            self.cos_cached[:, :, :seq_len, ...].to(dtype=dtype),
            self.sin_cached[:, :, :seq_len, ...].to(dtype=dtype),
        )

def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)

def apply_rotary_pos_emb(q, k, cos, sin, position_ids):
    # The first two dimensions of cos and sin are always 1, so we can `squeeze` them.
    cos = cos.squeeze(1).squeeze(0)  # [seq_len, dim]
    sin = sin.squeeze(1).squeeze(0)  # [seq_len, dim]
    cos = cos[position_ids].unsqueeze(1)  # [bs, 1, seq_len, dim]
    sin = sin[position_ids].unsqueeze(1)  # [bs, 1, seq_len, dim]
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed

# %% ../nbs/60_approximators/11_approximators.ipynb 9
class CausalSelfAttention(nn.Module):

    """ Causeal self-attention module
    Based on the implementation in https://github.com/time-series-foundation-models/lag-llama,
    without usage of kv_cache since we always make a prediction for only the next step
    """

    def __init__(self, n_embd_per_head, n_head, block_size, dropout) -> None:
        super().__init__()
        # query projections for all heads, but in a batch
        self.q_proj = nn.Linear(
            n_embd_per_head * n_head,
            n_embd_per_head * n_head,
            bias=False,
        )
        # key, value projections
        self.kv_proj = nn.Linear(
            n_embd_per_head * n_head,
            2 * n_embd_per_head * n_head,
            bias=False,
        )
        # output projection
        self.c_proj = nn.Linear(
            n_embd_per_head * n_head,
            n_embd_per_head * n_head,
            bias=False,
        )

        self.n_head = n_head
        self.n_embd_per_head = n_embd_per_head
        self.block_size = block_size
        self.dropout = dropout

        self.rope_scaling=None

        self._init_rope()

    def _init_rope(self):
        if self.rope_scaling is None:
            self.rotary_emb = LlamaRotaryEmbedding(
                self.n_embd_per_head, max_position_embeddings=self.block_size
            )

        else:
            raise NotImplementedError("RoPE scaling is not yet implemented")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # batch size, sequence length, embedding dimensionality (n_embd)

        B, T, C = x.size()

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q = self.q_proj(x)
        k, v = self.kv_proj(x).split(self.n_embd_per_head * self.n_head, dim=2)

        
        k = k.view(B, -1, self.n_head, self.n_embd_per_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, -1, self.n_head, self.n_embd_per_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, -1, self.n_head, self.n_embd_per_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        if self.rotary_emb is not None:
            cos, sin = self.rotary_emb(device=v.device, dtype=v.dtype, seq_len=T)
            q, k = apply_rotary_pos_emb(q, k, cos, sin, position_ids=None)


        y = F.scaled_dot_product_attention(
            q, k, v, attn_mask=None, dropout_p=self.dropout, is_causal=True
        )
        
        # # debug
        # if not torch.isfinite(y).all():
        #     print("y is not finite")
        #     print(y)
        #     print(q)
        #     print(k)
        #     print(v)

        # re-assemble all head outputs side by side
        y = y.transpose(1, 2).contiguous().view(B, T, C)

        # output projection
        y = self.c_proj(y)

        return y

def find_multiple(n: int, k: int) -> int:
    if n % k == 0:
        return n
    return n + k - (n % k)

# %% ../nbs/60_approximators/11_approximators.ipynb 10
class MLP_block(nn.Module):
    def __init__(self, n_embd_per_head, n_head, min_multiple = 256, gating = True) -> None:
        super().__init__()
        hidden_dim = 4 * n_embd_per_head * n_head
        n_hidden = int(2 * hidden_dim / 3)
        self.gating = gating
        
        n_hidden = find_multiple(n_hidden, min_multiple)

        self.c_fc1 = nn.Linear(
            n_embd_per_head * n_head, n_hidden, bias=False
        )
        if gating:
            self.c_fc2 = nn.Linear(
                n_embd_per_head * n_head, n_hidden, bias=False
            )
        
        self.c_proj = nn.Linear(
            n_hidden, n_embd_per_head * n_head, bias=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.gating:
            x = F.silu(self.c_fc1(x)) * self.c_fc2(x)
        else:
            x = F.silu(self.c_fc1(x))
        x = self.c_proj(x)
        return x

# %% ../nbs/60_approximators/11_approximators.ipynb 11
class RMSNorm(nn.Module):
    """Root Mean Square Layer Normalization as implemented in https://github.com/time-series-foundation-models/lag-llama.

    Derived from https://github.com/bzhangGo/rmsnorm/blob/master/rmsnorm_torch.py. BSD 3-Clause License:
    https://github.com/bzhangGo/rmsnorm/blob/master/LICENSE.
    """

    def __init__(self, size: int, dim: int = -1, eps: float = 1e-5) -> None:
        super().__init__()
        self.scale = nn.Parameter(torch.ones(size))
        self.eps = eps
        self.dim = dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        norm_x = x.to(torch.float32).pow(2).mean(dim=self.dim, keepdim=True)
        x_normed = x * torch.rsqrt(norm_x + self.eps)
        output = (self.scale * x_normed).type_as(x)
        return output

# %% ../nbs/60_approximators/11_approximators.ipynb 12
class Block(nn.Module):
    def __init__(self, n_embd_per_head, n_head, block_size, dropout, min_multiple = 256, gating=True) -> None:
        super().__init__()
        self.rms_1 = RMSNorm(n_embd_per_head * n_head)
        self.attn = CausalSelfAttention(n_embd_per_head, n_head, block_size, dropout)
        self.rms_2 = RMSNorm(n_embd_per_head * n_head)
        self.mlp = MLP_block(n_embd_per_head, n_head, min_multiple = min_multiple, gating=gating)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.attn(self.rms_1(x))
        y = x + self.mlp(self.rms_2(x))
        return y



"""Simple agents that can be used as a baseline or for testing purposes."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/40_base_agents/11_basic_agents.ipynb.

# %% auto 0
__all__ = ['RandomAgent']

# %% ../../nbs/40_base_agents/11_basic_agents.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, Optional, List
import numpy as np

from ..envs.base import BaseEnvironment
from .base import BaseAgent
from ..utils import MDPInfo

# %% ../../nbs/40_base_agents/11_basic_agents.ipynb 4
class RandomAgent(BaseAgent):

    """
    A random agent that samples actions from the environment's action space. Useful for testing and as minimal baseline.
    """

    def __init__(self,
                environment_info: MDPInfo,
                obsprocessors: list[object] | None = None,
                agent_name: str = "RandomAgent",
                *args,
                **kwargs,
                ):

        super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name)

    def draw_action_(self, observation):
        return self.environment_info.action_space.sample()

    def fit(self, X, Y):
        assert self.mode == "train"
        pass

    def save(self, path):
        pass

    def load(self, path):
        pass



"""Some general utility functions that are used throughout the package."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils/00_utils.ipynb.

# %% auto 0
__all__ = ['check_parameter_types', 'Parameter', 'MDPInfo', 'DatasetWrapper', 'DatasetWrapperMeta', 'merge_dictionaries',
           'set_param']

# %% ../nbs/00_utils/00_utils.ipynb 3
from torch.utils.data import Dataset
from typing import Union, List, Tuple, Literal
from gymnasium.spaces import Space
from .dataloaders.base import BaseDataLoader

import logging

import numpy as np

# %% ../nbs/00_utils/00_utils.ipynb 4
def check_parameter_types(
                            *args,                      # any number of parameters to be checked
                            parameter_type=np.ndarray   # the expected type for each parameter
):
    
    """
    Checks if each argument in args is of the specified type, defaulting to np.ndarray.
    """
    
    for index, arg in enumerate(args):
        if not isinstance(arg, parameter_type):
            raise TypeError(f"Argument {index+1} of {len(args)} is of type {type(arg).__name__}, expected {parameter_type.__name__}")

# %% ../nbs/00_utils/00_utils.ipynb 7
class Parameter():

    """
    Simple class to handle parameters in the environment. The advantage of this class is that it can be
    used to set parameters that may change over time and accessed by multiple objects such as the 
    environment, agent or dataloaders.
    """
    
    def __init__(self,
            value: int | float | list[int] | list[float] | np.ndarray,  # the value of the parameter  
            min_value: int | float | list[int] | list[float] | np.ndarray = None,  # the minimum value of the parameter
            max_value: int | float | list[int] | list[float] | np.ndarray = None,  # the maximum value of the parameter
            shape: tuple[int] = (1,)): # the shape of the parameter

        self._min_value = min_value
        self._max_value = max_value
        
        self.set_value(value, shape)

    def __call__(self):
        """
        Returns the parameter value. Alternative to get_value().

        """
        return self.get_value()

    def get_value(self):
        
        """
        Returns the parameter value.

        """

        return self._value

    def set_value(self, 
                    value: Union[int, float, List[int], List[float], np.ndarray], # the value of the parameter
                    shape: Tuple[int] = (1,)): # the shape of the parameter
       
        """
        Set the value of the parameter. The value can be a scalar, list or numpy array. The shape of the value
        must be the same as the shape of the parameter.

        All values, including scalars, are converted to numpy arrays. If the value is a scalar, it is reshaped
        to the shape of the parameter. If the value is a list or numpy array, it must have the same shape as the
        parameter. If the value is a scalar, it is reshaped to the shape of the parameter.

        Optionally, the value can be clipped to a minimum and maximum value.
        """

        if isinstance(value, (int, float)):
            self._value = np.array([value])
            self._value.reshape(shape)
        
        elif isinstance(value, list):
            value = np.array(value)
            assert value.shape == shape, "Shape of value must be the same as the shape of the parameter"
            self._value = value
        
        elif isinstance(value, np.ndarray):
            assert value.shape == shape, "Shape of value must be the same as the shape of the parameter"
            self._value = value
        
        else:
            raise ValueError("Value must be a scalar or numpy array")

        if self._min_value is not None:
            self._value = np.maximum(self._value, self._min_value)
        if self._max_value is not None:
            self._value = np.minimum(self._value, self._max_value)

    @property
    def shape(self):
        """
        Returns: The shape of the table of parameters.
        """
        return self._value.shape
    
    @property
    def size(self):
        """
        Returns: The size of the table of parameters.
        """
        return self._value.size 

# %% ../nbs/00_utils/00_utils.ipynb 16
class MDPInfo():
    """
    This class is used to store the information of the environment.
    It is based on MushroomRL (https://github.com/MushroomRL). It can be accessed by 
    agents that need the information of the environment, such as the state and action spaces.
    
    Key difference with MushroomRL is that the state and action spaces are gymnasium spaces.
    """
    
    def __init__(self,
                observation_space: Space,
                action_space: Space,  
                gamma: float,
                horizon: int,
                dt: float = 1e-1,
                backend: Literal['numpy'] = 'numpy'  # Currently only numpy is supported
            ) -> None: 

        self.observation_space = observation_space
        self.action_space = action_space
        self.gamma = gamma
        self.horizon = horizon
        self.dt = dt
        self.backend = backend

    @property
    def size(self):
        """
        Returns: The sum of the number of discrete states and discrete actions. Only works for discrete spaces.

        """
        return self.observation_space.size + self.action_space.size

    @property
    def shape(self):
        """
        Returns: The concatenation of the shape tuple of the state and action spaces.

        """
        return self.observation_space.shape + self.action_space.shape

# %% ../nbs/00_utils/00_utils.ipynb 20
class DatasetWrapper(Dataset):
    """
    This class is used to wrap a Pytorch Dataset around the ddopai dataloader
    to enable the usage of the Pytorch Dataloader during training. This way,
    agents that are trained using Pytorch without interacting with the environment
    can directly train on the data generated by the dataloader.
    
    """

    def __init__(self, 
            dataloader: BaseDataLoader, # Any dataloader that inherits from BaseDataLoader
            obsprocessors: List = None # processors (to mimic the environment processors)
            ):
        self.dataloader = dataloader
        self.obsprocessors = obsprocessors or []
    
    def __getitem__(self, idx):
        """
        Get the item at the provided idx.

        """

        # create tuple of items

        output = self.dataloader[idx]

        X = output[0]

        X = np.expand_dims(X, axis=0) # single datapoints are always returned without batch dimension, need to add for obsprocessors

        for obsprocessor in self.obsprocessors:
            X = obsprocessor(X)
        
        X = np.squeeze(X, axis=0) # remove batch dimension

        output = (X, *output[1:])
        
        return output


    def __len__(self):

        """

        Returns the length of the dataset. Depends on the state of
        the dataloader (train, val, test).

        """
        
        if self.dataloader.dataset_type == 'train':
            return self.dataloader.len_train
        elif self.dataloader.dataset_type == 'val':
            return self.dataloader.len_val
        elif self.dataloader.dataset_type == 'test':
            return self.dataloader.len_test
        else:
            raise ValueError("Dataset type must be either 'train', 'val' or 'test'")

# %% ../nbs/00_utils/00_utils.ipynb 24
class DatasetWrapperMeta(DatasetWrapper):
    """
    This class is used to wrap a Pytorch Dataset around the ddopai dataloader
    to enable the usage of the Pytorch Dataloader during training. This way,
    agents that are trained using Pytorch without interacting with the environment
    can directly train on the data generated by the dataloader.
    
    """

    def __init__(self, 
            dataloader: BaseDataLoader, # Any dataloader that inherits from BaseDataLoader
            draw_parameter_function: callable = None, # function to draw parameters from distribution
            distribution: Literal["fixed", "uniform"] | List = "fixed", # distribution for params during training, can be List for multiple parameters
            parameter_names: List[str] = None, # names of the parameters
            bounds_low: Union[int, float] | List = 0, # lower bound for params during training, can be List for multiple parameters
            bounds_high: Union[int, float] | List = 1, # upper bound for params during training, can be List for multiple parameters
            obsprocessors: List = None # processors (to mimic the environment processors)
            ):

        if isinstance(distribution, list) or isinstance(bounds_low, list) or isinstance(bounds_high, list):
            raise NotImplementedError("Multiple parameters not yet implemented")
        if obsprocessors is None:
            raise ValueError("Obsprocessors must be provided")
        
        self.distribution = [distribution]
        self.bounds_low = [bounds_low]
        self.bounds_high = [bounds_high]
        
        self.dataloader = dataloader

        self.draw_parameter = draw_parameter_function
        self.obsprocessors = obsprocessors

        self.parameter_names = parameter_names
    
    def __getitem__(self, idx):
        """
        Get the item at the provided idx.

        """

        features, demand = self.dataloader[idx] 

        features = np.expand_dims(features, axis=0) # add batch dimension as meta environments also return a batch dimension (needed for obsprocessor)

        params = {}
        for i in range(len(self.distribution)):
            param = self.draw_parameter(self.distribution[0], self.bounds_low[0], self.bounds_high[0], samples=1) # idx always gets a single sample
            params[self.parameter_names[i]] = param
        
        obs = params.copy()
        obs["features"] = features

        for obsprocessor in self.obsprocessors:
            obs = obsprocessor(obs)

        obs = np.squeeze(obs, axis=0) # remove batch dimension after observation has been processed as the pytorch dataloader adds the batch dimension

        return obs, demand, params

# %% ../nbs/00_utils/00_utils.ipynb 27
def merge_dictionaries(dict1, dict2):
    """ Merge two dictionaries. If a key is found in both dictionaries, raise a KeyError. """
    for key in dict2:
        if key in dict1:
            raise KeyError(f"Duplicate key found: {key}")
    
    # If no duplicates are found, merge the dictionaries
    merged_dict = {**dict1, **dict2}
    return merged_dict

# %% ../nbs/00_utils/00_utils.ipynb 29
def set_param(obj,
                name: str, # name of the parameter (will become the attribute name)
                input: Parameter | int | float | np.ndarray | List | None , # input value of the parameter
                shape: tuple = (1,), # shape of the parameter
                new: bool = False, # whether to create a new parameter or update an existing one
                ): 
    """
        Set a parameter for the class. It converts scalar values to numpy arrays and ensures that
        environment parameters are either of the Parameter class of Numpy arrays. If new is set to True, 
        the function will create a new parameter or update an existing one otherwise. If new is set to
        False, the function will raise an error if the parameter does not exist.
    """

    if input is None:
        param = None

    elif isinstance(input, Parameter):
        if input.shape != shape:
            raise ValueError("Parameter shape must be equal to the shape specified for this environment parameter")
        param = input
    
    elif isinstance(input, (int, float, np.integer, np.floating)):
        param = np.full(shape, input)

    elif isinstance(input, list):
        input = np.array(input)
        if input.shape == shape:
            param = input
        elif input.size == 1:  # Handle single-element arrays correctly
            param = np.full(shape, input.item())
        else:
            raise ValueError("Error in setting parameter. Input array must match the specified shape or be a single-element array")

    elif isinstance(input, np.ndarray):
        if input.shape == shape:
            param = input
        elif input.size == 1:  # Handle single-element arrays correctly
            param = np.full(shape, input.item())
        else:
            raise ValueError("Error in setting parameter. Input array must match the specified shape or be a single-element array")
    else:
        raise TypeError(f"Input must be a Parameter, scalar, or numpy array, got {type(input).__name__} with value {input}")

    # set the parameter
    if new:
        if hasattr(obj, name):
            logging.warning(f"Parameter {name} already exists. Overwriting it.")
        setattr(obj, name, param)
    else:
        if not hasattr(obj, name):
            raise AttributeError(f"Parameter {name} does not exist")
        else:
            setattr(obj, name, param)



"""Postprocessors can be used to process the output of an agent before it is being passed to the environment."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils/10_postprocessors.ipynb.

# %% auto 0
__all__ = ['ClipAction', 'RoundAction', 'MoveBatchToProductDim']

# %% ../nbs/00_utils/10_postprocessors.ipynb 3
from typing import Union, Optional

import numpy as np
from .utils import Parameter, check_parameter_types

import torch
import torch.nn as nn
import torch.nn.functional as F

# %% ../nbs/00_utils/10_postprocessors.ipynb 4
class ClipAction():
    """
    A class to clip input values within specified bounds.
    If the parameters lower and upper are not specified, no clipping is performed.
    If the parameters are scalar values, then all elements of the input are clipped to the same bounds.
    If the parameters are arrays, then each element of the input is clipped to the corresponding bounds.
    """

    def __init__(self, lower: Optional[Union[float, int, np.ndarray]] = None, upper: Optional[Union[float, np.ndarray]] = None): #
        self.lower = self._convert_to_array(lower)
        self.upper = self._convert_to_array(upper)

    def _convert_to_array(self, value: Optional[Union[float, int, list, np.ndarray]]) -> Optional[np.ndarray]:
        """
        Converts a float value to a numpy array of shape (1,) if needed.
        """
        if value is None:
            return None
        if isinstance(value, (float, int)):
            return np.array([value])
        if isinstance(value, list):
            return np.array(value)
        if isinstance(value, np.ndarray):
            return value
        raise TypeError(f"Bounds must be float or np.ndarray, got {type(value).__name__}")

    def __call__(self, input: np.ndarray) -> np.ndarray: #
        """
        Clips the input array within the specified bounds.
        """
        
        check_parameter_types(input)

        # Ensure bounds match the input's shape if they are arrays
        if self.lower is not None and self.lower.size != 1 and self.lower.shape != input.shape:
            raise ValueError("Lower bound array must match the input shape or be a single element")

        if self.upper is not None and self.upper.size != 1 and self.upper.shape != input.shape:
            raise ValueError("Upper bound array must match the input shape or be a single element")

        # Perform clipping
        output = np.clip(input, a_min=self.lower, a_max=self.upper)

        return output

# %% ../nbs/00_utils/10_postprocessors.ipynb 8
class RoundAction():
    """
    A class to round input values to the nearest specified unit size.
    Unit size can be any decimal value like 10, 3, 1, 0.1, 0.03, etc.
    """

    def __init__(self, unit_size: Union[float, int, np.ndarray]): #
        self.unit_size = self._validate_unit_size(unit_size)

    def _validate_unit_size(self, unit_size: Union[float, int, np.ndarray]) -> np.ndarray:
        """
        Ensures that the unit size is a positive float, int, or a numpy array of positive values.
        """
        if isinstance(unit_size, (float, int)):
            if unit_size <= 0:
                raise ValueError("Unit size must be a positive number")
            return np.array([unit_size], dtype=float)  # Convert to float for consistent behavior
        elif isinstance(unit_size, np.ndarray):
            if np.any(unit_size <= 0):
                raise ValueError("All elements of unit size array must be positive")
            return unit_size.astype(float)  # Ensure numpy array is of float type
        else:
            raise TypeError("Unit size must be a float, int, or np.ndarray")

    def __call__(self, input: np.ndarray) -> np.ndarray:
        
        """
        Rounds the input array to the nearest specified unit size.
        """
        
        check_parameter_types(input)

        # Ensure unit_size matches the input's shape if it is an array
        if self.unit_size.size != 1 and self.unit_size.shape != input.shape:
            raise ValueError("Unit size array must match the input shape or be a single element")

        # Perform rounding
        output = np.round(input / self.unit_size) * self.unit_size

        return output

# %% ../nbs/00_utils/10_postprocessors.ipynb 14
class MoveBatchToProductDim():
    """
    A class that moves the first dimension to the last place. Usefull for meta learners that 
    return the predictions of various units in the batch dimension while in environment the
    num_unit (e.g., num_SKU) dimension is usually the last one
    """

    def __init__(self,
                remove_action_per_unit_dim: bool = False # If there is only one action per unit, the action dimension can be removed by setting this to True
                ): 
        self.remove_action_per_unit_dim = remove_action_per_unit_dim

    def __call__(self, input: np.ndarray) -> np.ndarray:
        
        """
        Moves the first dimension to the last place.
        """

        if len(input.shape) == 1:
            raise ValueError("Input array must have at least 2 dimensions")

        output = np.moveaxis(input, 0, -1)

        if self.remove_action_per_unit_dim:
            if len(output.shape) > 2:
                raise ValueError("Removing action dim only works for arrays of shape (num_units, num_actions_per_unit)")
            if output.shape[0] == 1:
                output = np.squeeze(output, axis=0)
            else:
                raise ValueError("Removing action dim only works for arrays with one action per unit)")

        return output



"""Base environment with some basic funcitons"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/21_envs_inventory/10_base_inventory_env.ipynb.

# %% auto 0
__all__ = ['BaseInventoryEnv']

# %% ../../../nbs/21_envs_inventory/10_base_inventory_env.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, Tuple, List

from ..base import BaseEnvironment
from ...utils import Parameter, MDPInfo
from ...dataloaders.base import BaseDataLoader
from ...loss_functions import pinball_loss

import gymnasium as gym

import numpy as np
import time

# %% ../../../nbs/21_envs_inventory/10_base_inventory_env.ipynb 4
class BaseInventoryEnv(BaseEnvironment):
    """
    Base class for inventory management environments. This class inherits from BaseEnvironment.
    
    """

    def __init__(self, 

        ## Parameters for Base env:
        mdp_info: MDPInfo, #
        postprocessors: list[object] | None = None,  # default is empty list
        mode: str = "train", # Initial mode (train, val, test) of the environment
        return_truncation: str = True, # whether to return a truncated condition in step function
        dataloader: BaseDataLoader = None, # dataloader for the environment
        horizon_train: int = 100, # horizon for training mode
        
        # Parameters common in all inventory environments
        underage_cost: Union[np.ndarray, Parameter, int, float] = 1, # underage cost per unit
        overage_cost: Union[np.ndarray, Parameter, int, float] = 0, # overage cost per unit (zero in most cases)

        ) -> None:

        self.dataloader = dataloader
        
        self.set_param("underage_cost", underage_cost, shape=(self.num_SKUs[0],), new=True)
        self.set_param("overage_cost", overage_cost, shape=(self.num_SKUs[0],), new=True)

        super().__init__(mdp_info=mdp_info, postprocessors = postprocessors,  mode = mode, return_truncation=return_truncation, horizon_train=horizon_train)
    
    def set_observation_space(self,
                            shape: tuple, # shape of the dataloader features
                            low: Union[np.ndarray, float] = -np.inf, # lower bound of the observation space
                            high: Union[np.ndarray, float] = np.inf, # upper bound of the observation space
                            samples_dim_included = True # whether the first dimension of the shape input is the number of samples
                            ) -> None:
        
        '''
        Set the observation space of the environment.
        This is a standard function for simple observation spaces. For more complex observation spaces,
        this function should be overwritten. Note that it is assumped that the first dimension
        is n_samples that is not relevant for the observation space.

        '''

        # To handle cases when no external information is available (e.g., parametric NV)
        
        if shape is None:
            self.observation_space = None

        else:
            if not isinstance(shape, tuple):
                raise ValueError("Shape must be a tuple.")
            
            if samples_dim_included:
                shape = shape[1:] # assumed that the first dimension is the number of samples

            self.observation_space = gym.spaces.Box(low=low, high=high, shape=shape, dtype=np.float32)

    def set_action_space(self,
                            shape: tuple, # shape of the dataloader target
                            low: Union[np.ndarray, float] = -np.inf, # lower bound of the observation space
                            high: Union[np.ndarray, float] = np.inf, # upper bound of the observation space
                            samples_dim_included = True # whether the first dimension of the shape input is the number of samples
                            ) -> None:
        
        '''
        Set the action space of the environment.
        This is a standard function for simple action spaces. For more complex action spaces,
        this function should be overwritten. Note that it is assumped that the first dimension
        is n_samples that is not relevant for the action space.
        '''

        if not isinstance(shape, tuple):
            raise ValueError("Shape must be a tuple.")
        
        if samples_dim_included:
            shape = shape[1:] # assumed that the first dimension is the number of samples

        self.action_space = gym.spaces.Box(low=low, high=high, shape=shape, dtype=np.float32)
    
    def get_observation(self):
        
        """
        Return the current observation. This function is for the simple case where the observation
        is only an x,y pair. For more complex observations, this function should be overwritten.

        """

        X_item, Y_item = self.dataloader[self.index]

        return X_item, Y_item
    
    def reset(self,
        start_index: int | str = None, # index to start from
        state: np.ndarray = None # initial state
        ) -> Tuple[np.ndarray, bool]:

        """
        Reset function for the Newsvendor problem. It will return the first observation and demand.
        For val and test modes, it will by default reset to 0, while for the train mode it depends
        on the paramter "horizon_train" whether a random point in the training data is selected or 0
        """

        truncated = self.reset_index(start_index)

        observation, self.demand = self.get_observation()
        
        return observation




"""Very high-level functions to run experiments with minimal code, directly from terminal."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb.

# %% auto 0
__all__ = ['set_warnings', 'prep_experiment', 'init_wandb', 'track_libraries_and_git', 'import_config',
           'transfer_lag_window_to_env', 'transfer_additional_target_to_env', 'get_ddop_data', 'download_data',
           'set_indices', 'set_up_env', 'set_up_earlystoppinghandler', 'prep_and_run_test', 'clean_up', 'select_agent',
           'merge_with_namespace']

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, List, Tuple, Dict, Literal
import logging
from datetime import datetime  
import numpy as np
import sys
import os
import yaml
import pickle
import warnings
import torch

from .tracking import get_git_hash, get_library_version
from .agents.class_names import AGENT_CLASSES
from .dataloaders.tabular import XYDataLoader
from .datasets.default_datasets import DatasetLoader
from .experiment_functions import EarlyStoppingHandler, test_agent

import wandb

import gc

import importlib

from tqdm import tqdm, trange

# Think about how to handle mushroom integration.
from mushroom_rl.core import Core

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 5
def set_warnings (logging_level):

    """ Set warnings to be ignored for the given logging level or higher."""

    if logging.getLogger().isEnabledFor(logging_level):
        warnings.filterwarnings("ignore", category=UserWarning, message=".*Box bound precision lowered by casting to float32.*")
        warnings.filterwarnings("ignore", category=UserWarning, message=".*TypedStorage is deprecated.*")
        warnings.filterwarnings("ignore", category=FutureWarning, message=".*You are using `torch.load` with `weights_only=False`.*")

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 7
def prep_experiment(
    project_name: str,
    libraries_to_track: List[str] = ["ddopai"],
    config_train_name: str = "config_train",
    config_agent_name: str = "config_agent",
    config_env_name: str = "config_env",
):
    """ First stpes to always execute when starting an experiment (using wandb for tracking)"""

    init_wandb(project_name)
    track_libraries_and_git(libraries_to_track)

    config_train = import_config(config_train_name)
    config_agent = import_config(config_agent_name) # General config file containing all agent parameters
    config_env = import_config(config_env_name)

    AgentClass = select_agent(config_train["agent"]) # Select agent class and import dynamically
    agent_name = config_train["agent"]
    config_agent = config_agent[config_train["agent"]] # Get parameters for specific agent
    
    transfer_lag_window_to_env(config_env, config_agent) 

    wandb.config.update(config_train)
    wandb.config.update(config_agent)
    wandb.config.update(config_env)

    return config_train, config_agent, config_env, AgentClass, agent_name

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 8
def init_wandb(project_name: str): #

    """ init wandb """

    wandb.init(
        project=project_name,
        name = f"{project_name}_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    )

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 9
def track_libraries_and_git(    libraries_to_track: List[str],
                                tracking: bool = True,
                                tracking_tool = "wandb", # Currenty only wandb is supported
                                ) -> None:
    
    """
    Track the versions of the libraries and the git hash of the repository.

    """

    for lib in libraries_to_track:
        version = get_library_version(lib, tracking=tracking, tracking_tool=tracking_tool)
        logging.info(f"{lib}: {version}")
    git_hash = get_git_hash(".", tracking=tracking, tracking_tool=tracking_tool)
    logging.info(f"Git hash: {git_hash}")

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 10
def import_config(  filename: str, # Name of the file, must be a yaml file
                    path: str = None # Optional path to the file if it is not in the current directory
                    ) -> Dict:
    
    """
    Import a config file in YAML format

    """

    # Check if filename has an extension
    if '.' in filename:
        extension = filename.split(".")[-1]
    else:
        extension = ''

    if not extension:
        filename += ".yaml"
    elif extension not in ["yaml", "yml"]:
        raise ValueError("The configuration file must have a .yaml or .yml extension.")


    if path is not None:
        full_path = os.path.join(path, filename)
    else:
        full_path = filename
    

    # Check if the file exists
    if not os.path.exists(full_path):
        raise FileNotFoundError(f"The configuration file '{full_path}' does not exist.")

    with open(full_path, "r") as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise yaml.YAMLError(f"Error parsing YAML file '{full_path}': {exc}")
    
    logging.info(f"Configuration file '{filename}' successfully loaded.")
    logging.debug(f"Configuration: {config}")

    return config

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 11
def transfer_lag_window_to_env(config_env: Dict, #
                                config_agent: Dict
                                ) -> None:
    
    """
    Transfer the lag window from the agent configuration to the environment configuration

    """
    
    if "lag_window" in config_agent.keys():
        if isinstance(config_agent["lag_window"], int):
            config_env["lag_window_params"]["lag_window"] = config_agent["lag_window"]
        else:
            raise ValueError("The lag window must be an integer.")
        del config_agent["lag_window"]
    else:
        logging.warning("No lag window specified in the agent configuration. Keeping value from env config")

#| export
def transfer_additional_target_to_env(config_env: Dict, #
                                config_agent: Dict
                                ) -> None:
    
    """
    Transfer the lag window from the agent configuration to the environment configuration

    """
    
    if "provide_additional_target" in config_agent.keys():
        if isinstance(config_agent["provide_additional_target"], bool):
            config_env["provide_additional_target"] = config_agent["provide_additional_target"]
        else:
            raise ValueError(f"Expected boolean value for provide_additional_target, got {config_agent['provide_additional_target']}")
        del config_agent["provide_additional_target"]

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 13
def get_ddop_data(
    config_env: Dict,
    overwrite: bool = False
    ) -> Tuple:

    """ Standard function to load data provided by the ddop package """
    
    data = download_data(config_env, overwrite)

    val_index_start, test_index_start = set_indices(config_env, data[0])

    return data, val_index_start, test_index_start


# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 14
def download_data(  config_env: Dict,
                    overwrite: bool = False #
                    ) -> Tuple:

    """ Download standard dataset from ddop repository using the DatasetLoader class """

    datasetloader = DatasetLoader()

    data = datasetloader.load_dataset(
        dataset_type = config_env["dataset_type"],
        dataset_number = config_env["dataset_number"],
        overwrite=False)

    data_tuple = data["data_raw_features"], data["data_raw_target"]

    return data_tuple

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 15
def set_indices(config_env: Dict, #
                X: np.ndarray 
) -> Tuple:

    """ Set the indices for the validation and test set """

    val_index_start = len(X) - config_env["size_val"] - config_env["size_test"]
    test_index_start = len(X) - config_env["size_test"]

    return val_index_start, test_index_start

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 17
def set_up_env(
    env_class,
    raw_data: Tuple, #
    val_index_start: int,
    test_index_start: int,
    config_env: Dict,
    postprocessors: List,
) -> object:
    
    """ Set up the environment """

    dataloader = XYDataLoader(  X = raw_data[0],
                                Y = raw_data[1],
                                val_index_start = val_index_start,
                                test_index_start = test_index_start,
                                lag_window_params = config_env["lag_window_params"],
                                normalize_features = {'normalize': config_env["normalize_features"], 'ignore_one_hot': True})

    environment = env_class(
        dataloader = dataloader,
        postprocessors = postprocessors,
        **config_env["env_kwargs"]
    )

    return environment

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 19
def set_up_earlystoppinghandler(config_train: Dict) -> object: #

    """ Set up the early stopping handler """

    # check if config_train has either early_stopping_patience or early_stopping_warmup
    if "early_stopping_patience" in config_train or "early_stopping_warmup" in config_train:
        warmup = config_train["early_stopping_warmup"] if "early_stopping_warmup" in config_train else 0
        patience = config_train["early_stopping_patience"] if "early_stopping_patience" in config_train else 0

        earlystoppinghandler = EarlyStoppingHandler(warmup=warmup, patience=patience)
    else:
        earlystoppinghandler = None

    return earlystoppinghandler

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 21
def prep_and_run_test(
    agent,
    environment,
    agent_dir: str = None,
    save_dataset: bool = True,
    save_features: bool = False,
    dataset_dir: str = None,
    eval_step_info = False,
    tracking = "wandb"):
    
    """
    Test the agent in the environment.
    """

    if save_dataset:
        if dataset_dir is None:
            raise ValueError("If save_dataset is True, dataset_dir must be specified.")

    # load parameters of agent

    if agent_dir is not None:
        agent.load(agent_dir)

    # Set agent and environment to test mode
    agent.eval()
    environment.test()

    # Run test episode
    output = test_agent(
        agent,
        environment,
        return_dataset=save_dataset,
        save_features=save_features,
        tracking=tracking,
        eval_step_info=eval_step_info,
    )

    # Save dataset
    if save_dataset:

        R, J, dataset = output

        if not os.path.exists(dataset_dir):
            os.mkdir(dataset_dir)
        else:
            raise ValueError("Path to save dataset already exists") # it should never exist since run_id is usually part or path and unique
        
        dir = os.path.join(dataset_dir, "dataset_test.pkl")

        with open (os.path.join(dir), "wb") as f:
            pickle.dump(dataset, f)

        artifact = wandb.Artifact("transition_test_set", type="dataset")

        artifact.add_file(os.path.join(dir))

        wandb.run.log_artifact(artifact)
    
    else:

        R, J = output

    logging.info(f"final evaluation on test set: R = {np.round(R, 10)} J = {np.round(J, 10)}")



# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 23
def clean_up(agent, environment):

    """ Clean up agent and environment to free up GPU memory """
    
    # Delete agent and environment to free up GPU memory
    del agent
    del environment

    # Force garbage collection
    gc.collect()

    # Clear GPU cache
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    wandb.finish()

    return None, None

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 25
def select_agent(agent_name: str) -> type: #
    """ Select an agent class from a list of agent names and return the class"""
    if agent_name in AGENT_CLASSES:
        module_path, class_name = AGENT_CLASSES[agent_name].rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    else:
        raise ValueError(f"Unknown agent name: {agent_name}")

# %% ../nbs/30_experiment_functions/20_meta_experiment_functions.ipynb 26
def merge_with_namespace(target_dict, source_dict, target_dict_name):
    
    """
    Merge source_dict into target_dict, using the keys as namespaces.
    For example, if target_dict_name is "agent", the key "agent-epsilon" in source_dict will be merged into target_dict["epsilon"].
    The function is to merge hyperparameters from a config file with the default hyperparameters from the yaml files

    Args:
        target_dict (dict): Target dictionary
        source_dict (dict): Source dictionary
        target_dict_name (str): Name of the target dictionary

    Returns:
        dict: Merged dictionary

    """
    
    for namespaced_key, value in source_dict.items():
        keys = namespaced_key.split('-')

        if keys[0] != target_dict_name:
            continue

        keys = keys[1:]

        d = target_dict
    
        # Check if keys already exist in target_dict
        exists = True
        for key in keys:
            if key not in d:
                exists = False
                break
            if isinstance(d[key], dict):
                d = d[key]
                continue
            else:
                break

        # If all keys are present, overwrite the value
        if exists:
            prev = d[key]
            d[key] = value
            print(f"Overwriting in key {namespaced_key} value {prev} with value {value}")
        else:
            # exception if key is not present in target_dict
            print(f"Key {namespaced_key} not found in {target_dict_name}.")
            raise ValueError(f"Key {namespaced_key} not found in {target_dict_name}.")

    return target_dict



"""Base agent that all agents shall inherit from"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/40_base_agents/10_base_agents.ipynb.

# %% auto 0
__all__ = ['BaseAgent']

# %% ../../nbs/40_base_agents/10_base_agents.ipynb 3
# import logging
# logging_level = logging.DEBUG

from abc import ABC, abstractmethod
from typing import Union, Optional, List, Tuple
import numpy as np

from ..envs.base import BaseEnvironment
from ..utils import MDPInfo, Parameter
import numbers

# # TEMPORARY
# from sklearn.utils.validation import check_array
# import numbers

# %% ../../nbs/40_base_agents/10_base_agents.ipynb 4
class BaseAgent():

    """  
    Base class for all agents to enforce a common interface. See below for more detailed description of the requriements.

    """

    train_mode = "direct_fit" # or "epochs_fit" or "env_interaction"
    
    def __init__(self,
                    environment_info: MDPInfo,
                    obsprocessors: list[object] | None = None,  # default is empty list
                    agent_name: str | None = None,
                    receive_batch_dim: bool = False
                 ):

        self.obsprocessors = obsprocessors or []

        self.environment_info = environment_info
        self.mode = "train"
        self.print = False  # Can be used for debugging
        self.receive_batch_dim = receive_batch_dim

        self.agent_name = agent_name

    def draw_action(self, observation: np.ndarray) -> np.ndarray: #

        """
        Main interfrace to the environemnt. Applies preprocessors to the observation.
        Internal logic of the agent to be implemented in draw_action_ method.
        """

        batch_added = False
        if not isinstance(observation, dict):
            observation = self.add_batch_dim(observation) # adds batch dim if self.receive_batch_dim is False
            batch_added = True

        for obsprocessor in self.obsprocessors:
            observation = obsprocessor(observation) # applies all preprocessors to the dict observation
            if not isinstance(observation, dict) and not batch_added: # checks if one of the processors has removed the dict structure
                observation = self.add_batch_dim(observation) # adds batch dim afterwards, if self.receive_batch_dim is False    
                batch_added = True

        action = self.draw_action_(observation)

        return action

    @abstractmethod
    def draw_action_(self, observation: np.ndarray) -> np.ndarray: #
        """Generate an action based on the observation - this is the core method that needs to be implemented by all agents."""
        pass

    def add_obsprocessor(self, obsprocessor: object): # pre-processor object that can be called via the "__call__" method
        """Add a preprocessor to the agent"""
        self.obsprocessors.append(obsprocessor)

    def train(self):
        """Set the internal state of the agent to train"""
        self.mode = "train"
        
    def eval(self):
        """
        Set the internal state of the agent to eval. Note that for agents we do not differentiate between val and test modes.

        """
        self.mode = "eval"
    
    def add_batch_dim(self, input: np.ndarray) -> np.ndarray: #
        
        """
        Add a batch dimension to the input array if it doesn't already have one.
        This is necessary because most environments do not have a batch dimension, but agents typically expect one.
        If the environment does have a batch dimension, the agent can set the receive_batch_dim attribute to True to skip this step.

        """

        if self.receive_batch_dim:
            # If the batch dimension is expected, return the input as is
            return input
        else:
            # Add a batch dimension by expanding the dimensions of the input
            return np.expand_dims(input, axis=0)
        
    def save(self):
        """Save the agent's parameters to a file."""
        raise NotImplementedError("This agent does not have a save method implemented.")

    def load(self):
        """Load the agent's parameters from a file."""
        raise NotImplementedError("This agent does not have a load method implemented.")

    @staticmethod
    def update_model_params(default_params: dict, custom_params: dict) -> dict: #
        """ override default parameters with custom parameters in a dictionary"""
        updated_params = default_params.copy()
        updated_params.update(custom_params)
        return updated_params

    @staticmethod
    def convert_to_numpy_array(
        input: np.ndarray | List | float | int | Parameter | None #
    ):

        """convert input to numpy array or keep as Parameter"""

        if isinstance(input, np.ndarray):
            return input
        elif isinstance(input, (list, numbers.Number)):
            return np.array(input)
        elif isinstance(input, Parameter):
            return input
        else:
            raise ValueError("Input type not recognized.")

    @staticmethod
    def convert_recursively_to_int(
        input = List | Tuple, #
    ) -> List | Tuple:

        """convert all elements of a list or tuple to int"""

        if isinstance(input, list):
            return [BaseAgent.convert_recursively_to_int(item) for item in input]
        elif isinstance(input, tuple):
            return tuple(BaseAgent.convert_recursively_to_int(item) for item in input)
        else:
            # If it's not a list or tuple, convert to int directly
            return int(input)
